# Copyright 2016 Enzo Busseti, Stephen Boyd, Steven Diamond, BlackRock Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module defines cost models for both simulation and optimization.

We implement two types of costs, as discussed in the paper:
:class:`TransactionCost`, defined in :paper:`section 2.3 <section.2.3>`, and
:class:`HoldingCost`, defined in :paper:`section 2.4 <section.2.4>`. We also
provide versions of each that are specialized to the stock market,
:class:`StocksTransactionCost` and :class:`StocksHoldingCost`: these have a
default values that are typical for liquid stocks in the US.

The latter two are included by default in
:class:`cvxportfolio.StockMarketSimulator`, the market simulator specialized to
the stock market, which is used throughout the :doc:`examples <examples>`. So,
the example back-tests all include these costs, unless otherwise specified.

The cost objects have the same user interface when you use them as simulator
costs or optimization costs. For example, to include an annualized
borrowing cost of 10% (as opposed to the default 5%) in a back-test, you do

.. code-block:: python

    simulator = cvx.MarketSimulator(
        universe, costs=[cvx.HoldingCost(short_fees=10)])
    backtest_result_with_borrow_fee = simulator.backtest(policy)

And to include the same borrow cost penalty in an optimization-based policy,
you do

.. code-block:: python

    policy = cvx.SinglePeriodOptimization(
        objective = cvx.ReturnsForecast()
            - 0.5 * cvx.FullCovariance()
            - cvx.HoldingCost(short_fees=10),
        constraints = [cvx.LeverageLimit(3)])

As done throughout the library, you can pass data in the form of a Python
scalar (like 10), a Pandas Series indexed by time, by the assets, or a
Pandas DataFrame indexed by time and with the assets as columns; see the
:ref:`manual page on passing data <passing-data>`.
"""

import copy
import warnings
from numbers import Number

import cvxpy as cp
import numpy as np
import pandas as pd

from .constraints.base_constraints import (CostInequalityConstraint,
                                           EqualityConstraint,
                                           InequalityConstraint)
from .errors import ConvexityError, ConvexSpecificationError
from .estimator import (CvxpyExpressionEstimator, DataEstimator, Estimator,
                        SimulatorEstimator)
from .forecast import HistoricalMeanVolume, HistoricalStandardDeviation
from .hyperparameters import HyperParameter, _resolve_hyperpar
from .utils import periods_per_year_from_datetime_index, resample_returns

__all__ = ["HoldingCost", "TransactionCost", "SoftConstraint",
           "StocksTransactionCost", "StocksHoldingCost", "TcostModel",
           "HcostModel"]


class Cost(CvxpyExpressionEstimator): # pylint: disable=abstract-method
    """Base class for cost objects (and also risks).

    You should derive from this class to define an objective term for
    optimization-based policies, like a risk model.

    The base class itself defines logic used for the algebraic operations, and
    to take inequalities of a cost object (which results in a Cvxportfolio
    :doc:`constraint <constraints>` object).
    """

    def __mul__(self, other):
        """Multiply by constant."""
        if isinstance(other, (Number, HyperParameter)):
            return CombinedCosts([self], [other])
        raise SyntaxError(
            "You can only multiply a cost instance by a scalar "
            + "or a HyperParameter instance.")

    def __add__(self, other):
        """Add cost expression to another cost expression.

        Idea is to create a new CombinedCost class that
        implements `compile_to_cvxpy` and values_in_time_recursive
        by summing over costs.
        """
        if isinstance(other, CombinedCosts):
            return other.__radd__(self)
        return CombinedCosts([self, other], [1.0, 1.0])

    def __rmul__(self, other):
        """Multiply by constant."""
        return self * other

    def __radd__(self, other):
        """Add cost expression to another cost."""
        return self + other

    def __neg__(self):
        """Take negative of cost."""
        return self * -1

    def __sub__(self, other):
        """Subtract other cost."""
        return self.__add__(-other)

    def __rsub__(self, other):
        """Subtract from other cost."""
        return other.__add__(-self)

    def __le__(self, other):
        """Self <= other, return CostInequalityConstraint.

        For now we check here the type of "other" but it would be nicer
        to have CostInequalityConstraint's internal DataEstimator throw
        NotImplemented instead.
        """
        if isinstance(other, (Number, pd.Series)):
            return CostInequalityConstraint(self, other)
        return NotImplemented

    def __lt__(self, other):
        """Self < other."""
        raise SyntaxError(
            'Strict inequalities are not allowed in convex programs.')

    def __gt__(self, other):
        """Self > other."""
        raise SyntaxError(
            'Strict inequalities are not allowed in convex programs.')

    def __ge__(self, other):
        """Self >= other, return CostInequalityConstraint."""
        return (-self).__le__(other)


class CombinedCosts(Cost):
    """Algebraic combination of :class:`Cost` instances.

    :param costs: instances of :class:`Cost`
    :type costs: list
    :param multipliers: floats that multiply the ``costs``
    :type multipliers: list
    """

    def __init__(self, costs, multipliers):
        for cost in costs:
            if not isinstance(cost, Cost):
                raise SyntaxError(
                    "You can only sum cost instances to other cost instances.")
        self.costs = costs
        self.multipliers = multipliers
        # this is changed by WorstCaseRisk before compiling to Cvxpy
        # TODO this can be problematic, also CostConstraint needs to act on it
        self.do_convexity_check = True

    def __add__(self, other):
        """Add self to other (combined) cost."""
        if isinstance(other, CombinedCosts):
            return CombinedCosts(self.costs + other.costs,
                                 self.multipliers + other.multipliers)
        return CombinedCosts(self.costs + [other], self.multipliers + [1.0])

    def __radd__(self, other):
        """Add other (combined) cost to self."""
        if isinstance(other, CombinedCosts):
            return other + self # pragma: no cover
        return CombinedCosts([other] + self.costs, [1.0] + self.multipliers)

    def __mul__(self, other):
        """Multiply by constant."""
        return CombinedCosts(self.costs,
                             [el * other for el in self.multipliers])

    def __neg__(self):
        """Take negative of cost."""
        return self * -1

    def initialize_estimator_recursive(self, **kwargs):
        """Initialize iterating over constituent costs.

        :param kwargs:  All parameters passed to :meth:`initialize_estimator`.
        :type kwargs: dict
        """
        for el in self.costs:
            el.initialize_estimator_recursive(**kwargs)

    def finalize_estimator_recursive(self, **kwargs):
        """Finalize iterating over constituent costs.

        :param kwargs:  All parameters passed to :meth:`finalize_estimator`.
        :type kwargs: dict
        """
        for el in self.costs:
            el.finalize_estimator_recursive(**kwargs)

    def values_in_time_recursive(self, **kwargs):
        """Evaluate estimators by iterating over constituent costs.

        :param kwargs: All parameters passed to :meth:`values_in_time`.
        :type kwargs: dict
        """
        for el in self.costs:
            el.values_in_time_recursive(**kwargs)

    def compile_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):
        """Compile cost by iterating over constituent costs.

        :param w_plus: Post-trade weights.
        :type w_plus: cvxpy.Variable
        :param z: Trade weights.
        :type z: cvxpy.Variable
        :param w_plus_minus_w_bm: Post-trade weights minus benchmark weights.
        :type w_plus_minus_w_bm: cvxpy.Variable

        :raises cvxportfolio.errors.ConvexSpecificationError: If the compiled
            Cvxpy expression doesn't follow the disciplined convex programming
            rules.
        :raises cvxportfolio.errors.ConvexityError: If the compiled Cvxpy
            expression is not concave (since we maximize it).

        :returns: Cvxpy expression of the combined cost.
        :rtype: cvxpy.Expression
        """
        expression = 0
        for multiplier, cost in zip(self.multipliers, self.costs):
            add = _resolve_hyperpar(multiplier) *\
                    cost.compile_to_cvxpy(w_plus, z, w_plus_minus_w_bm)
            if not add.is_dcp():
                raise ConvexSpecificationError(cost * multiplier)
            if self.do_convexity_check and (not add.is_concave()):
                raise ConvexityError(cost * multiplier)
            if (not self.do_convexity_check) and add.is_concave():
                raise ConvexityError(-cost * multiplier)
            expression += add
        return expression

    def collect_hyperparameters(self):
        """Collect hyper-parameters in the combined cost.

        :returns: List of :class:`cvxportfolio.hyperparameters.HyperParameter`
            instances.
        :rtype: list
        """
        return sum([el.collect_hyperparameters() for el in self.costs], []) +\
            sum([el.collect_hyperparameters() for el in self.multipliers if
                hasattr(el, 'collect_hyperparameters')], [])

    def __repr__(self):
        """Pretty-print."""
        result = ''
        for i, (mult, cost) in enumerate(zip(self.multipliers, self.costs)):
            if not isinstance(mult, HyperParameter):
                if mult == 0:
                    continue
                if mult < 0:
                    result += ' - ' if i > 0 else '-'
                else:
                    result += ' + ' if i > 0 else ''
                result += (str(abs(mult)) + ' * ' if abs(mult) != 1 else '')
            else:
                result += str(mult) + ' * '
            result += cost.__repr__()
        return result

    def _copy_keeping_multipliers(self):
        """This method is used when creating MPO policies.

        We want to deepcopy the constituent cost objects, but not the
        multipliers (which can be symbolic HPs).

        We will probably re-factor this, or make it public.
        """
        return CombinedCosts(
            costs = [el._copy_keeping_multipliers()
                if hasattr(el, '_copy_keeping_multipliers')
                    else copy.deepcopy(el)
                        for el in self.costs],
            multipliers = self.multipliers)


class SoftConstraint(Cost):
    """Soft constraint cost.

    This can be applied to most :doc:`constraint objects <constraints>`,
    as discussed in :ref:`its section of the constraints documentation
    <soft-constraints>`.

    :param constraint: Cvxportfolio constraint instance whose violation
        we penalize.
    :type constraint: cvxportfolio.constraints.EqualityConstraint or
        cvxportfolio.constraints.InequalityConstraint
    """

    def __init__(self, constraint):
        self.constraint = constraint

    def compile_to_cvxpy( # pylint: disable=inconsistent-return-statements
        self, w_plus, z, w_plus_minus_w_bm):
        """Compile cost to cvxpy expression.

        :param w_plus: Post-trade weights.
        :type w_plus: cvxpy.Variable
        :param z: Trade weights.
        :type z: cvxpy.Variable
        :param w_plus_minus_w_bm: Post-trade weights minus benchmark weights.
        :type w_plus_minus_w_bm: cvxpy.Variable

        :raises SyntaxError: If the constraint is not a EqualityConstraint or
             InequalityConstraint.

        :return: Cvxpy expression of the soft constraint.
        :rtype: cvxpy.Expression
        """

        try:
            expr = (self.constraint._compile_constr_to_cvxpy(
                w_plus, z, w_plus_minus_w_bm) - self.constraint._rhs())

        except AttributeError as exc:
            raise SyntaxError(
                f"{self.__class__.__name__} can only be used with"
                " EqualityConstraint or InequalityConstraint instances."
                    ) from exc

        if isinstance(self.constraint, EqualityConstraint):
            return cp.sum(cp.abs(expr))
        if isinstance(self.constraint, InequalityConstraint):
            return cp.sum(cp.pos(expr))


def _annual_percent_to_per_period(value, ppy):
    """Transform annual percent to per-period return.

    :param value: Annual percent return.
    :type value: float
    :param ppy: Periods per year.
    :type ppy: int

    :returns: Per-period return.
    :rtype: float
    """
    return resample_returns(returns=value/100, periods=ppy)


class SimulatorCost( # pylint: disable=abstract-method
    SimulatorEstimator, Cost):
    """Cost class that can be used by :class:`cvxportfolio.MarketSimulator`.

    This is the base class of both :class:`HoldingCost` and
    :class:`TransactionCost`.

    This class derives from :class:`Cost` and
    :class:`cvxportfolio.estimator.SimulatorEstimator`.
    It implements the :meth:`simulate` method (which is abstract in
    :class:`cvxportfolio.estimator.SimulatorEstimator`). The implementation
    uses the CVXPY compiled expression of the (optimization) cost to evaluate
    the cost in simulation, so we're sure the algebra is (exactly) the same.
    Of course the CVXPY expression operates on weights, so holdings and trades
    are divided by the portfolio value, and the result is multiplied by the
    portfolio value. If you implement a custom simulator's cost and prefer
    to implement the cost expression directly you can derive straight from
    :class:`cvxportfolio.estimator.SimulatorEstimator`. Look at the
    code of the cost classes we implement if in doubt. Every operation that
    is different in simulation and in optimization is wrapped in a
    :class:`cvxportfolio.estimator.SimulatorEstimator` which implements
    different :meth:`values_in_time` and :meth:`simulate` respectively.
    """

    def initialize_estimator( # pylint: disable=arguments-differ
        self, universe, **kwargs):
        """Initialize cost by compiling its CVXPY expression (if applies).

        This must be called by derived classes if you want to use
        an internal CVXPY expression to evaluate the simulator cost.

        :param universe: Current trading universe.
        :type universe: pd.Index
        :param kwargs: Other unused arguments to :meth:`initialize_estimator`.
        :type kwargs: dict
        """
        # pylint: disable=attribute-defined-outside-init
        if hasattr(self, 'compile_to_cvxpy'):
            self._w_plus = cp.Variable(len(universe))
            self._z = cp.Variable(len(universe))
            self._w_plus_minus_w_bm = cp.Variable(len(universe))
            self._cvxpy_expression = self.compile_to_cvxpy(
                w_plus = self._w_plus, z = self._z,
                w_plus_minus_w_bm = self._w_plus_minus_w_bm)

    def simulate( # pylint: disable=arguments-differ,too-many-arguments
        self, t, u, h_plus, past_volumes,
        past_returns, current_prices,
        current_weights, current_portfolio_value, **kwargs):
        """Simulate the cost in the market simulator (not optimization).

        Cost classes that are meant to be used in the simulator can
        implement this. The arguments to this are the same as for
        :meth:`cvxportfolio.estimator.Estimator.values_in_time` plus the
        trades vector and post-trade allocations.

        :param t: Current timestamp.
        :type t: pandas.Timestamp
        :param u: Trade vector in cash units requested by the policy.
            If the market simulator implements rounding by number of shares
            and/or canceling trades on assets whose volume for the period
            is zero, this is after those transformations.
        :type u: pandas.Series
        :param h_plus: Post-trade holdings vector.
        :type h_plus: pandas.Series
        :param past_returns: Past market returns (including cash).
        :type past_returns: pandas.DataFrame
        :param past_volumes: Past market volumes, or None if not available.
        :type past_volumes: pandas.DataFrame or None
        :param current_prices: Current (open) prices, or None if not available.
        :type current_prices: pandas.Series or None
        :param current_weights: Current allocation weights (before trading).
        :type current_weights: pandas.Series
        :param current_portfolio_value: Current total value of the portfolio
            in cash units, before costs.
        :type current_portfolio_value: float
        :param kwargs: Other unused arguments to
            :meth:`cvxportfolio.estimator.SimulatorEstimator.simulate`.
        :type kwargs: dict

        :returns: Simulated realized value of the cost, in cash accounting
            units (e.g., US dollars). Typically a positive number: it is
            subtracted from the cash account.
        :rtype: float
        """

        self.values_in_time(
            t=t, past_volumes=past_volumes, past_returns=past_returns,
            current_prices=current_prices, current_weights=current_weights,
            current_portfolio_value=current_portfolio_value)

        self._w_plus.value = h_plus.values / current_portfolio_value
        self._z.value = u.values / current_portfolio_value
        return self._cvxpy_expression.value * current_portfolio_value


class YearDividedByTradingPeriod(SimulatorEstimator):
    """Length of a year divided by this trading period's.

    This is used by :class:`HoldingCost` to model its separate behaviors in
    optimization and simulation.

    :param periods_per_year: If provided, overrides internal estimation of
        number of periods per year in optimization. Has no effect in
        simulation. Default None.
    :type periods_per_year: int or None
    """

    def __init__(self, periods_per_year=None):
        self.periods_per_year = periods_per_year

    def values_in_time( # pylint: disable=arguments-differ
            self, past_returns, **kwargs):
        """Evaluate in optimization.

        :param past_returns: Past market returns, we use its index to estimate
            the typical length of a trading period.
        :type past_returns: pd.DataFrame
        :param kwargs: Other unused arguments to
            :meth:`Estimator.values_in_time`.
        :type kwargs: dict

        :returns: Trading periods per year.
        :rtype: float
        """
        if self.periods_per_year is None:
            return periods_per_year_from_datetime_index(past_returns.index)
        return self.periods_per_year

    def simulate( # pylint: disable=arguments-differ
            self, t, t_next, **kwargs):
        """Evaluate in simulation.

        We use the real time length of the trading period, which we know.

        :param t: Current timestamp.
        :type t: pd.Timestamp
        :param t_next: Timestamp of next trading period.
        :type t_next: pd.Timestamp
        :param kwargs: Other unused arguments to
            :meth:`SimulatorEstimator.simulate`.
        :type kwargs: dict

        :returns: Trading periods per year.
        :rtype: float
        """
        return pd.Timedelta('365.24d') / (t_next - t)


class HoldingCost(SimulatorCost):
    r"""Generic holding cost model.

    This is a generalization of the model described in :paper:`section 2.4
    <section.2.4>` of the paper (which instead corresponds to
    :class:`StocksHoldingCost`).

    This represents the following objective term,
    expressed in terms of the post-trade dollar positions:

    .. math::

        s^T_t {(h^+_t)}_- + l^T_t {(h^+_t)}_+ - d^T_t h^+_t

    where :math:`s_t` are the (short) borrowing fees,
    :math:`l_t` are the fees on long positions,
    and :math:`d_t` are dividend rates (their sign is flipped because
    the costs are deducted from the cash account at each period).

    :param short_fees: Short borrowing fees expressed as annual percentage;
        you can provide them as a float (constant for all times and all
        assets), a :class:`pd.Series` indexed by time (constant for all
        assets but varying in time) or by assets' names (constant in time
        but varying across assets), or a :class:`pd.DataFrame` indexed by
        time and whose columns are the assets' names, if varying both
        in time and across assets. See :ref:`the manual page on passing data
        <passing-data>`. If `None` (the default) the term is
        ignored.
    :type short_fees: float, pd.Series, pd.DataFrame or None
    :param long_fees: Fees on long positions expressed as annual percentage;
        same convention as above applies.
    :type long_fees: float, pd.Series, pd.DataFrame or None
    :param dividends: Dividend rates per period. Same conventions as above.
        Our default data interface already includes the dividend rates in the
        market returns (*i.e.*, uses total returns, from the adjusted
        prices). If, however, you provide your own market returns that do not
        include dividends, you may use this. Default None, which disables the
        term.
    :type dividends: float, pd.Series, pd.DataFrame or None
    :param periods_per_year: Number of trading period per year, used to obtain
        the holding cost per-period from the annualized percentages. Only
        relevant in optimization, since in simulation we know the exact
        duration of each period. If left to the default, None, uses the
        estimated average length of each period from the historical data.
        Note that, in simulation, the holding cost is applied to the actual
        length of the period between trading times, so for example it will
        be higher over a weekend than between weekdays.
    :type periods_per_year: float or None
    """

    def __init__(self, short_fees=None, long_fees=None, dividends=None,
                 periods_per_year=None):

        self.short_fees = None if short_fees is None else DataEstimator(
            short_fees)
        self.long_fees = None if long_fees is None else DataEstimator(
            long_fees)
        self.dividends = None if dividends is None else DataEstimator(
            dividends)
        self.periods_per_year = YearDividedByTradingPeriod(periods_per_year)
        self._short_fees_parameter = None
        self._long_fees_parameter = None
        self._dividends_parameter = None

    def initialize_estimator( # pylint: disable=arguments-differ
            self, universe, **kwargs):
        """Initialize cvxpy parameters.

        We don't use the parameter from
        :class:`cvxportfolio.estimator.DataEstimator` because we need to
        divide the value by periods_per_year.

        :param universe: Trading universe, including cash.
        :type universe: pandas.Index
        :param kwargs: Other unused arguments to :meth:`initialize_estimator`.
        :type kwargs: dict
        """

        if self.short_fees is not None:
            self._short_fees_parameter = cp.Parameter(len(universe) - 1,
                nonneg=True)

        if self.long_fees is not None:
            self._long_fees_parameter = cp.Parameter(len(universe) - 1,
                nonneg=True)

        if self.dividends is not None:
            self._dividends_parameter = cp.Parameter(len(universe) - 1)

        # Used by SimulatorCost
        super().initialize_estimator(universe=universe, **kwargs)

    def values_in_time( # pylint: disable=arguments-differ
            self, **kwargs):
        """Update cvxpy parameters.

        We compute the estimate of periods per year from past returns
        (if not provided by the user) and populate the cvxpy parameters
        with the current values of the user-provided data, transformed
        to per-period.

        :param kwargs: Other unused arguments to :meth:`values_in_time`.
        :type kwargs: dict
        """

        if self.short_fees is not None:
            self._short_fees_parameter.value = np.ones(
                self._short_fees_parameter.size
                ) * _annual_percent_to_per_period(
                    self.short_fees.current_value,
                    self.periods_per_year.current_value)

        if self.long_fees is not None:
            self._long_fees_parameter.value = np.ones(
                self._long_fees_parameter.size
                ) * _annual_percent_to_per_period(
                    self.long_fees.current_value,
                    self.periods_per_year.current_value)

        if self.dividends is not None:
            self._dividends_parameter.value =\
                np.ones(self._dividends_parameter.size
                ) * self.dividends.current_value

    def compile_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):
        """Compile cost to cvxpy expression.

        :param w_plus: Post-trade weights.
        :type w_plus: cvxpy.Variable
        :param z: Trade weights.
        :type z: cvxpy.Variable
        :param w_plus_minus_w_bm: Post-trade weights minus benchmark
            weights.
        :type w_plus_minus_w_bm: cvxpy.Variable

        :returns: Cvxpy expression.
        :rtype: cvxpy.expression
        """

        expression = 0.

        if self.short_fees is not None:
            expression += self._short_fees_parameter.T @ cp.neg(w_plus[:-1])

        if self.long_fees is not None:
            expression += self._long_fees_parameter.T @ cp.pos(w_plus[:-1])

        if self.dividends is not None:
            # we have a minus sign because costs are deducted from PnL
            expression -= self._dividends_parameter.T @ w_plus[:-1]

        assert expression.is_convex()
        return expression


class StocksHoldingCost(HoldingCost):
    r"""Holding cost specialized to stocks (only borrow fee).

    This implements the simple model described in :paper:`section 2.4
    <section.2.4>` of the paper, refer to that for more details. The cost is,
    in terms of the post-trade dollar positions:

    .. math::

        s^T_t {(h^+_t)}_-,

    *i.e.*, a simple borrow fee applies to the short positions. This class is a
    simplified version of :class:`HoldingCost`: it drops the ``long_fees`` and
    ``dividends`` parameters and keeps only ``short_fees`` with a default
    value of 5, *i.e.*, :math:`5\%` annualized borrowing fee. That is a rough
    (optimistic) approximation of the cost of shorting liquid US stocks. This
    cost is included **by default** in :class:`StockMarketSimulator`, the
    market simulator specialized to US (liquid) stocks.

    :param short_fees: Same as in :class:`HoldingCost`: annualized borrow fee
        in percent, can be asset- and/or period-specific. Default 5.
    :type short_fees: float, pd.Series, pd.DataFrame or None
    """

    def __init__(self, short_fees=5):
        super().__init__(short_fees=short_fees)


class VolumeHatOrRealized(SimulatorEstimator):
    r"""Predictor of market volumes used by :class:`TransactionCost`.

    This is used to model the different behaviors in optimization and
    simulation. (In the latter case we use the realized volumes.)

    :param volume_hat: Estimator of the :math:`\hat V` in optimization. Unused
        in simulation.
    :type volume_hat: cvx.estimator.Estimator
    """

    def __init__(self, volume_hat):
        self.volume_hat = volume_hat

    def values_in_time( # pylint: disable=arguments-differ
            self, **kwargs):
        """Evaluate in optimization.

        :param kwargs: All arguments to
            :meth:`estimator.Estimator.values_in_time`.
        :type kwargs: dict

        :returns: Current estimate of market volumes.
        :rtype: np.ndarray
        """
        return self.volume_hat.current_value

    def simulate( # pylint: disable=arguments-differ
            self, current_volumes, **kwargs):
        """Evaluate in simulation.

        :param current_volumes: Current market volumes.
        :type current_volumes: pd.Series
        :param kwargs: All other arguments to
            :meth:`estimator.SimulatorEstimator.simulate`.
        :type kwargs: dict

        :raises SyntaxError: If the market volumes are not present in the
            market data.

        :returns: Current market volumes.
        :rtype: np.ndarray
        """
        assert self.volume_hat.current_value is None
        if current_volumes is None:
            raise SyntaxError(
                "If you don't provide volumes you should set b to None"
                " in the market simulator's TransactionCost object.")
        return current_volumes.values

class TransactionCost(SimulatorCost):
    r"""This is a generic model for transaction cost of financial assets.

    .. versionadded:: 1.2.0

        This was significantly improved with new options; it was undocumented
        before.

    It is described in :paper:`section 2.3 <section.2.3>` of the paper.

    The model is, separated on a single asset (equation 2.2 in the paper)

    .. math ::

        a | x | + b  \sigma \frac{{ | x |}^{3/2}}{V^{1/2}} + c x

    where :math:`x` is the dollar traded quantity,
    :math:`a` is a coefficient representing fees proportional to the absolute
    value traded, like half the bid-ask spread,
    :math:`b` is a coefficient that multiplies the market impact term,
    typically of the order of 1,
    :math:`\sigma` is an estimate of the volatility of the asset returns over
    recent periods,
    :math:`V` is the market volume traded over the period for the asset
    and :math:`c` is a coefficient used to introduce bias in the model,
    for example the negative of open-to-close return (if transactions are
    executed at close), or the negative of the open-to-VWAP return (if
    transactions are executed at the volume-weighted average price).

    In optimization the realized market volume :math:`V` is not known and
    we use its forecast :math:`\hat V` instead.

    As done throughout the library, this implementation accepts either
    :ref:`user-provided data <passing-data>` for the various parts of the
    model, or uses built-in :doc:`forecaster classes <forecasts>` to do the
    heavy-lifting.

    :param a: Coefficients of the first term of the transaction cost model, for
        example half the bid-ask spread, brokerage fees proportional to the
        size of each trade, .... :ref:`Usual conventions on passing data
        <passing-data>` apply. Default None, which disables the term.
    :type a: float, pd.Series, pd.DataFrame, or None
    :param b: Coefficients of the second term of the transaction cost model,
        typically of the order of 1. Same conventions. Default None, which
        disables the term and invalidates the following three parameters.
    :type b: float, pd.Series, pd.DataFrame, or None
    :param volume_hat: Forecast of the market volumes, has only effect in
        optimization (in simulation the actual volumes are used). You can
        pass a DataFrame of externally computed forecasts, or use the default
        :class:`cvxportfolio.forecast.HistoricalMeanVolume` (or another
        forecaster) to compute the forecasts at each point in time. If you pass
        a class, like the default, it is instantiated with parameter
        ``rolling`` equal to 1 year, if you prefer a different estimation
        lenght you can instantiate the forecaster and pass the instance.
    :type volume_hat: float, pd.Series, pd.DataFrame, cvx.forecast.BaseForecast
        class or instance
    :param sigma: Externally computed forecasts, or forecaster, of the market
        returns' volatilities :math:`\sigma`. The default is
        :class:`cvxportfolio.forecast.HistoricalStandardDeviation`. If you
        pass a class, like the default, it is instantiated with parameter
        ``rolling`` equal to 1 year, if you prefer a different estimation
        lenght you can instantiate the forecaster and pass the instance.
    :type sigma: float, pd.Series, pd.DataFrame, cvx.forecast.BaseForecast
        class or instance
    :param exponent: Exponent of the second term of the model, default
        :math:`3/2`. (In the model above, this exponent applies to
        :math:`|z|`, and this exponent minus 1 applies to the denominator
        term :math:`V`). You can use any float larger than 1.
    :type exponent: float
    :param c: Coefficients of the third term of the transaction cost model.
        If None, the default, the term is ignored.
    :type c: float, pd.Series, pd.DataFrame, or None
    """

    def __init__( # pylint: disable=too-many-arguments
        self, a=0., b=None, volume_hat=HistoricalMeanVolume,
        sigma=HistoricalStandardDeviation, exponent=1.5, c=None):

        self.a = None if a is None else DataEstimator(a)
        self.b = None if b is None else DataEstimator(b)
        self.c = None if c is None else DataEstimator(
            c, compile_parameter=True)

        if self.b is not None:
            if isinstance(volume_hat, type):
                volume_hat = volume_hat(
                    rolling=pd.Timedelta('365.24d'))
            self.market_volumes = VolumeHatOrRealized(DataEstimator(volume_hat))

            if isinstance(sigma, type):
                sigma = sigma(
                    rolling=pd.Timedelta('365.24d'))
            self.sigma = DataEstimator(sigma)

        if exponent < 1.:
            raise SyntaxError(
                'Exponent should be >=1, otherwise the'
                ' transaction cost model is not convex.')
        self.exponent = exponent
        self._first_term_multiplier = None
        self._second_term_multiplier = None

    def initialize_estimator( # pylint: disable=arguments-differ
            self, universe, **kwargs):
        """Initialize cvxpy parameters.

        :param universe: Trading universe, including cash.
        :type universe: pandas.Index
        :param kwargs: Other unused arguments to :meth:`initialize_estimator`.
        :type kwargs: dict
        """
        if self.a is not None:
            self._first_term_multiplier = cp.Parameter(
                len(universe)-1, nonneg=True)
        if self.b is not None:
            self._second_term_multiplier = cp.Parameter(
                len(universe)-1, nonneg=True)

        # SimulatorCost
        super().initialize_estimator(universe=universe, **kwargs)

    def values_in_time( # pylint: disable=arguments-differ
            self, current_portfolio_value, **kwargs):
        """Update cvxpy parameters.

        :param current_portfolio_value: Current total value of the portfolio.
        :type current_portfolio_value: float
        :param kwargs: Other unused arguments to :meth:`values_in_time`.
        :type kwargs: dict
        """
        if self.a is not None:
            self._first_term_multiplier.value = np.ones(
                self._first_term_multiplier.size) * self.a.current_value

        if self.b is not None:

            self._second_term_multiplier.value =\
                (self.b.current_value * self.sigma.current_value
                ) / ((self.market_volumes.current_value + 1E-8)
                    / current_portfolio_value) ** (self.exponent - 1)

    def compile_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):
        """Compile cost to cvxpy expression.

        :param w_plus: Post-trade weights.
        :type w_plus: cvxpy.Variable
        :param z: Trade weights.
        :type z: cvxpy.Variable
        :param w_plus_minus_w_bm: Post-trade weights minus benchmark
            weights.
        :type w_plus_minus_w_bm: cvxpy.Variable

        :returns: Cvxpy expression.
        :rtype: cvxpy.expression
        """
        expression = 0
        if self.a is not None:
            expression += cp.abs(z[:-1]) @ self._first_term_multiplier
            assert expression.is_convex()
        if self.b is not None:
            expression += (
                cp.abs(z[:-1]) ** (self.exponent)
                    ) @ self._second_term_multiplier
            assert expression.is_convex()
        if self.c is not None:
            expression += cp.sum(z[:-1] * self.c.parameter)
        return expression

# Backward compatibility before 1.2.0

class SimpleSigmaEst(SimulatorEstimator):
    """Simple estimator of sigma for backward compatibility."""

    def __init__(self, window_sigma_est):
        warnings.warn(
            "Passing a number to window_sigma_est is deprecated, "
            "You should use a forecaster like the default "
            "HistoricalStandardDeviation, instantiating with"
            " its 'rolling' argument to choose the length of estimation.",
            DeprecationWarning)
        self.window_sigma_est = window_sigma_est

    def values_in_time(
        # pylint: disable=arguments-differ
        self, past_returns, **kwargs):
        """Compute historical sigma.

        :param past_returns: Past market returns.
        :type past_returns: pd.DataFrame
        :param kwargs: Other unused arguments.
        :type kwargs: dict

        :returns: Estimated sigma
        :rtype: np.array
        """

        return np.sqrt(
            (past_returns.iloc[-self.window_sigma_est:, :-1]**2).mean()).values

    def simulate(
        # pylint: disable=arguments-differ
        self, current_returns, past_returns, **kwargs):
        """Compute historical sigma.

        :param current_returns: Current market returns.
        :type current_returns: pd.Series
        :param past_returns: Past market returns.
        :type past_returns: pd.DataFrame
        :param kwargs: Other unused arguments.
        :type kwargs: dict

        :returns: Estimated sigma
        :rtype: np.array
        """

        return np.std(pd.concat(
            [past_returns.iloc[-self.window_sigma_est + 1:, :-1],
            pd.DataFrame(current_returns.iloc[:-1]).T], axis=0), axis=0).values

        # sum = (past_returns.iloc[-self.window_sigma_est-1:, :-1]**2).sum()
        # count = past_returns.iloc[-self.window_sigma_est-1:, :-1].count()
        # sum += current_returns**2
        # count += ~current_returns.isnull()
        # return np.sqrt(sum / count).values

# Backward compatibility before 1.2.0

class SimpleVolumeEst(Estimator):
    """Simple estimator of volume for backward compatibility."""

    def __init__(self, window_volume_est):
        warnings.warn(
            "Passing a number to window_volume_est is deprecated, "
            "You should use a forecaster like the default "
            "HistoricalMeanVolume, instantiating with"
            " its 'rolling' argument to choose the length of estimation.",
            DeprecationWarning)

        self.window_volume_est = window_volume_est

    def values_in_time(
        # pylint: disable=arguments-differ
        self, past_volumes, **kwargs):
        """Compute historical mean of volumes.

        :param past_volumes: Past market volumes
        :type past_volumes: pd.DataFrame
        :param kwargs: Other unused arguments.
        :type kwargs: dict

        :raises SyntaxError: If the market data does not contain volumes.

        :returns: Estimated volume
        :rtype: np.array
        """

        if past_volumes is None:
            raise SyntaxError(
                "If you don't provide market volumes you can not use the"
                " market impact term of the transaction cost model.")

        return past_volumes.iloc[-self.window_volume_est:].mean().values

class StocksTransactionCost(TransactionCost):
    """Simplified version of :class:`TransactionCost` for stocks.

    .. versionadded:: 1.2.0

        We added the ``sigma`` and ``volume_hat`` parameters to support
        user-provided values as well as forecasters with various parameters.
        We deprecated the old way (``window_sigma_est`` and
        ``window_volume_est``) in which that was done before.

    This is included as a simulator cost by default (with default arguments) in
    :class:`cvxportfolio.StockMarketSimulator`.

    :param a: Same as in :class:`TransactionCost`, default 0.
    :type a: float or pd.Series or pd.DataFrame
    :param pershare_cost: Per-share cost: cash paid for each share traded.
        Requires to know the prices of the stocks (they are present in the
        default market data server
        :class:`cvxportfolio.data.DownloadedMarketData`). Default 0.005.
    :type pershare_cost: float or pd.Series or pd.DataFrame
    :param b: Same as in :class:`TransactionCost`, default 1.
    :type b: float or pd.Series or pd.DataFrame
    :param sigma: Same as in :class:`TransactionCost`.
    :type sigma: float, pd.Series, pd.DataFrame, cvx.forecast.BaseForecast
        class or instance
    :param window_sigma_est: Deprecated, size of rolling window to
        estimate historical standard deviations. If left to None (the default)
        has no effect. Use instead the ``sigma`` parameter.
    :type window_sigma_est: int or None
    :param volume_hat: Same as in :class:`TransactionCost`.
    :type volume_hat: float, pd.Series, pd.DataFrame, cvx.forecast.BaseForecast
        class or instance
    :param window_volume_est: Deprecated, size of rolling window to estimate
        the mean of past volumes. If left to None (the default)
        has no effect. Use instead the ``volume_hat`` parameter.
    :type window_volume_est: int or None
    :param exponent: Same as in :class:`TransactionCost`.
    :type exponent: float
    :param c: Same as in :class:`TransactionCost`.
    :type c: float or pd.Series or pd.DataFrame or None
    """

    def __init__( # pylint: disable=too-many-arguments
        self, a=0., pershare_cost=0.005, b=1.0,
        volume_hat=HistoricalMeanVolume, sigma=HistoricalStandardDeviation,
        exponent=1.5, c=None, window_sigma_est=None, window_volume_est=None):

        if window_sigma_est is not None:
            sigma = SimpleSigmaEst(window_sigma_est)

        if window_volume_est is not None:
            volume_hat = SimpleVolumeEst(window_volume_est)

        super().__init__(# because we update it with pershare_cost
                         a= 0. if a is None else a,
                         b=b, c=c, exponent=exponent,
                         volume_hat=volume_hat,
                         sigma=sigma,
                         )

        self.pershare_cost = DataEstimator(pershare_cost)\
            if pershare_cost is not None else None

    def values_in_time( # pylint: disable=arguments-renamed
            self, current_prices, **kwargs):
        """Update linear cost with per-share cost.

        :param current_prices: Current (open) market prices.
        :type current_prices: pd.Series
        :param kwargs: Other unused arguments to
            :meth:`cvxportfolio.estimator.Estimator.values_in_time`.
        :type kwargs: dict

        :raises SyntaxError: If market prices are not available and
            pershare_cost is not None.
        """

        super().values_in_time(current_prices=current_prices, **kwargs)

        if self.pershare_cost is not None:
            if current_prices is None:
                raise SyntaxError("If the market data doesn't contain prices"
                                  " you should set pershare_cost to None")
            assert not np.any(current_prices.isnull())
            assert not np.any(current_prices == 0.)
            self._first_term_multiplier.value += \
                self.pershare_cost.current_value / current_prices.values

# Aliases

class TcostModel(TransactionCost):
    """Alias of :class:`TransactionCost`.

    As it was defined originally in :paper:`section 6.1 <section.6.1>` of the
    paper.
    """

class HcostModel(HoldingCost):
    """Alias of :class:`HoldingCost`.

    As it was defined originally in :paper:`section 6.1 <section.6.1>` of the
    paper.
    """



# Copyright 2016 Enzo Busseti, Stephen Boyd, Steven Diamond, BlackRock Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module contains classes that define return models for portfolio
optimization policies and related objects."""

import cvxpy as cp
import numpy as np

from .costs import Cost
from .estimator import DataEstimator  # , ParameterEstimator
from .forecast import HistoricalMeanError, HistoricalMeanReturn

__all__ = [
    "ReturnsForecast",
    "ReturnsForecastError",
    "CashReturn",
]


class CashReturn(Cost):
    r"""Objective term representing cash return.

    By default, the forecast of cash return :math:`{\left(\hat{r}_t\right)}_n`
    is the observed value from last period :math:`{\left({r}_{t-1}\right)}_n`.

    This object is included automatically in
    :class:`cvxportfolio.SinglePeriodOptimization` and
    :class:`cvxportfolio.MultiPeriodOptimization` policies. You can change
    this behavior by setting their ``include_cash_return`` argument to False.
    If you do so, you may include this cost explicitely in the objective. You
    need to do so (only) if you provide your own cash return forecast.

    :param cash_returns: if you have your forecast for the cash return, you
        should pass it here, either as a float (if constant) or as pd.Series
        with datetime index (if it changes in time). If you leave the default,
        None, the cash return forecast at time t is the observed cash return
        at time t-1.
    :type cash_returns: float or pd.Series or None
    """

    def __init__(self, cash_returns=None):
        self.cash_returns = None if cash_returns is None else DataEstimator(
            cash_returns, compile_parameter=True)
        self._cash_return_parameter = None

    def initialize_estimator( # pylint: disable=arguments-differ
            self, **kwargs):
        """Initialize model.

        :param kwargs: Unused arguments to :meth:`initialize_estimator`.
        :type kwargs: dict
        """
        self._cash_return_parameter = (cp.Parameter()
            if self.cash_returns is None else self.cash_returns.parameter)

    def values_in_time( # pylint: disable=arguments-differ
            self, past_returns, **kwargs):
        """Update cash return parameter as last cash return.

        :param past_returns: Past market returns.
        :type past_returns: pandas.DataFrame
        :param kwargs: All other parameters to :meth:`values_in_time`.
        :type kwargs: dict
        """
        if self.cash_returns is None:
            self._cash_return_parameter.value = past_returns.iloc[-1, -1]

    def compile_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):
        """Compile to cvxpy expression.

        :param w_plus: Post-trade weights.
        :type w_plus: cvxpy.Variable
        :param z: Trade weights.
        :type z: cvxpy.Variable
        :param w_plus_minus_w_bm: Post-trade weights minus benchmark
            weights.
        :type w_plus_minus_w_bm: cvxpy.Variable

        :returns: Cvxpy expression representing the risk model.
        :rtype: cvxpy.expression
        """
        return w_plus[-1] * self._cash_return_parameter


class ReturnsForecast(Cost):
    r"""Returns forecast for non-cash assets, provided by the user or computed
    from the data.

    It represents the objective term:

    .. math::

        \hat{r}_t^T z_t.

    :ref:`Optimization-based policies` use this, typically as the first
    element of their objectives. See chapters 4 and 5 of the paper, for example
    :paper:`page 25 <section.4.1>` for more details.

    This class can either get return forecasts from the user, for example
    computed with some machine learning technique, or it can estimate them
    automatically from the data. The returns' forecast provided by the user 
    must be supplied for all assets excluding cash. See the :ref:`passing-data`
    manual page for more information on how these are passed.


    :param r_hat: constant or time varying returns estimates, either
        user-provided (see :ref:`the manual page on passing data
        <passing-data>` or computed internally by a forecaster class or
        instance. The default is
        :class:`cvxportfolio.forecast.HistoricalMeanReturn` which computes the
        historical means of the past returns, at each point in the back-test.
        It is instantiated internally with default parameters, so it computes
        the full historical means at each point in time. If you prefer
        to change that (e.g., do rolling mean or exponential moving window)
        you can instantiate :class:`cvxportfolio.forecast.HistoricalMeanReturn`
        with your chosen parameters and pass the instance.
    :type r_hat: pd.Series, pd.DataFrame, float,
        :class:`cvxportfolio.forecast.BaseForecast` class or instance
    :param decay: decay factor used in 
        :class:`cvxportfolio.MultiPeriodOptimization` policies. It is as a 
        number in :math:`[0,1]`. At step :math:`\tau` of the MPO policy, where 
        :math:`\tau=0` is the initial one, the return predictions are
        multiplied by :math:`\texttt{decay}^{\tau}`. So, ``decay`` close to
        zero models a `fast` signal while ``decay`` close to one a `slow`
        signal. The default value is 1.    
    :type decay: float



    :Example:

    >>> import cvxportfolio as cvx
    >>> policy = cvx.SinglePeriodOptimization(cvx.ReturnsForecast() - \
        0.5 * cvx.FullCovariance(), [cvx.LongOnly(), cvx.LeverageLimit(1)])
    >>> cvx.MarketSimulator(['AAPL', 'MSFT', 'GOOG']).backtest(policy).plot()

    Defines a single period optimization policy where the returns' forecasts
    :math:`\hat{r}_t` are the full average of past returns at each point in 
    time and the risk model is the full covariance, also computed from the past 
    returns.

    :Example:

    >>> my_forecasts = pd.DataFrame(...)
    >>> returns_model = cvx.ReturnsForecast(r_hat=my_forecasts)

    With user-provided forecasts.

    :Example:

    >>> from cvxportfolio.forecast import HistoricalMeanReturn
    >>> returns_model = cvx.ReturnsForecast(
            r_hat=HistoricalMeanReturn(rolling = pd.Timedelta('365d')))

    Instead of the default (full historical means at each point in time),
    use rolling means of 1 year.
    """

    def __init__(self, r_hat=HistoricalMeanReturn, decay=1.):

        if isinstance(r_hat, type):
            r_hat = r_hat()

        # we don't use DataEstimator's parameter
        # because we apply the decay
        self.r_hat = DataEstimator(r_hat)
        self.decay = decay
        self._r_hat_parameter = None

    def initialize_estimator( # pylint: disable=arguments-differ
            self, universe, **kwargs):
        """Initialize model with universe size.

        :param universe: Trading universe, including cash.
        :type universe: pandas.Index
        :param kwargs: Other unused arguments to :meth:`initialize_estimator`.
        :type kwargs: dict
        """
        self._r_hat_parameter = cp.Parameter(len(universe)-1)

    def values_in_time( # pylint: disable=arguments-differ
            self, mpo_step=0, **kwargs):
        """Update returns parameter knowing which MPO step we're at.

        :param mpo_step: MPO step, 0 is current.
        :type mpo_step: int
        :param kwargs: All other parameters to :meth:`values_in_time`.
        :type kwargs: dict
        """
        self._r_hat_parameter.value = \
            np.ones(self._r_hat_parameter.size) * \
            self.r_hat.current_value * self.decay**(mpo_step)

    def compile_to_cvxpy(self,  w_plus, z, w_plus_minus_w_bm):
        """Compile to cvxpy expression.

        :param w_plus: Post-trade weights.
        :type w_plus: cvxpy.Variable
        :param z: Trade weights.
        :type z: cvxpy.Variable
        :param w_plus_minus_w_bm: Post-trade weights minus benchmark
            weights.
        :type w_plus_minus_w_bm: cvxpy.Variable

        :returns: Cvxpy expression representing the risk model.
        :rtype: cvxpy.expression
        """
        return w_plus[:-1].T @ self._r_hat_parameter


class ReturnsForecastError(Cost):
    r"""Simple return forecast error risk with values provided by the user.

    It represents the objective term:

    .. math::

        \delta^T |w^+_t - w^\text{b}_t |

    Implements the model described in :paper:`chapter 4, page 31 <section.4.3>`
    of the paper. You can pass the penalization parameters (see
    the :ref:`passing-data` manual page) or rely on a forecaster to do so.
    The default is :class:`cvxportfolio.forecast.HistoricalMeanError`, which
    computes the standard deviation of the mean for each asset's past returns,
    iteratively during a back-test.

    :param deltas_errors: Constant per-symbol errors on the returns
        forecasts (if Series), or varying in time (if DataFrame), or
        fitted from the data as the standard deviation of the historical
        mean estimator.
    :type deltas_errors: pd.DataFrame or pd.Series
        or :class:`cvxportfolio.estimator.Estimator`:
    """

    def __init__(self, deltas=HistoricalMeanError):

        if isinstance(deltas, type):
            deltas = deltas()
        self.deltas = DataEstimator(deltas)
        self._deltas_parameter = None

    def initialize_estimator( # pylint: disable=arguments-differ
            self, universe, **kwargs):
        """Initialize model with universe size.

        :param universe: Trading universe, including cash.
        :type universe: pandas.Index
        :param kwargs: Unused arguments to :meth:`initialize_estimator`.
        :type kwargs: pandas.DatetimeIndex
        """
        self._deltas_parameter = cp.Parameter(len(universe)-1, nonneg=True)

    def values_in_time( # pylint: disable=arguments-differ
            self, **kwargs):
        """Update returns forecast error parameters.

        :param kwargs: All arguments to :meth:`values_in_time`.
        :type kwargs: dict
        """
        self._deltas_parameter.value = self.deltas.current_value

    def compile_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):
        """Compile to cvxpy expression.

        :param w_plus: Post-trade weights.
        :type w_plus: cvxpy.Variable
        :param z: Trade weights.
        :type z: cvxpy.Variable
        :param w_plus_minus_w_bm: Post-trade weights minus benchmark
            weights.
        :type w_plus_minus_w_bm: cvxpy.Variable

        :returns: Cvxpy expression representing the risk model.
        :rtype: cvxpy.expression
        """
        return cp.abs(w_plus_minus_w_bm[:-1]).T @ self._deltas_parameter



# Copyright 2023 Enzo Busseti
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Here we define many realistic constraints that apply to :ref:`portfolio
optimization trading policies <optimization-policies-page>`.

Some of them, like :class:`LongOnly`, are
very simple to use. Some others are more advanced,
for example :class:`FactorNeutral`
takes time-varying factor exposures as parameters.

For a minimal example we present the classic Markowitz allocation.

.. code-block:: python

    import cvxportfolio as cvx

    objective = cvx.ReturnsForecast() - gamma_risk * cvx.FullCovariance()

    # the policy takes a list of constraint instances
    constraints = [cvx.LongOnly(applies_to_cash=True)]

    policy = cvx.SinglePeriodOptimization(objective, constraints)
    print(cvx.MarketSimulator(universe).backtest(policy))

With this, we require that the optimal post-trade weights
found by the single-period optimization policy are non-negative.
In our formulation the full portfolio weights vector (which includes
the cash account) sums to one,
see equation :math:`(4.9)` at page 43 of
`the book <https://stanford.edu/~boyd/papers/pdf/cvx_portfolio.pdf>`_.
"""

from .base_constraints import *
from .constraints import *



# Copyright 2023 Enzo Busseti
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module defines base constraint classes."""


from ..estimator import CvxpyExpressionEstimator, DataEstimator

__all__ = ['Constraint', 'EqualityConstraint', 'InequalityConstraint']

class Constraint(CvxpyExpressionEstimator):
    """Base cvxpy constraint class."""

    def compile_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):
        """Compile constraint to cvxpy.

        :param w_plus: Post-trade weights.
        :type w_plus: cvxpy.Variable
        :param z: Trade weights.
        :type z: cvxpy.Variable
        :param w_plus_minus_w_bm: Post-trade weights minus benchmark
            weights.
        :type w_plus_minus_w_bm: cvxpy.Variable
        :returns: some cvxpy.constraints object, or list of those
        :rtype: cvxpy.constraints, list
        """
        raise NotImplementedError # pragma: no cover


class EqualityConstraint(Constraint):
    """Base class for equality constraints.

    This class is not exposed to the user, each equality
    constraint inherits from this and overrides the
    :func:`InequalityConstraint._compile_constr_to_cvxpy` and
    :func:`InequalityConstraint._rhs` methods.

    We factor this code in order to streamline the
    design of :class:`SoftConstraint` costs.
    """

    def compile_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):
        """Compile constraint to cvxpy.

        :param w_plus: Post-trade weights.
        :type w_plus: cvxpy.Variable
        :param z: Trade weights.
        :type z: cvxpy.Variable
        :param w_plus_minus_w_bm: Post-trade weights minus benchmark
            weights.
        :type w_plus_minus_w_bm: cvxpy.Variable
        :returns: Cvxpy constraints object.
        :rtype: cvxpy.constraints
        """
        return self._compile_constr_to_cvxpy(w_plus, z, w_plus_minus_w_bm) ==\
            self._rhs()

    def _compile_constr_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):
        """Cvxpy expression of the left-hand side of the constraint."""
        raise NotImplementedError # pragma: no cover

    def _rhs(self):
        """Cvxpy expression of the right-hand side of the constraint."""
        raise NotImplementedError # pragma: no cover


class InequalityConstraint(Constraint):
    """Base class for inequality constraints.

    This class is not exposed to the user, each inequality
    constraint inherits from this and overrides the
    :func:`InequalityConstraint._compile_constr_to_cvxpy` and
    :func:`InequalityConstraint._rhs` methods.

    We factor this code in order to streamline the
    design of :class:`SoftConstraint` costs.
    """

    def compile_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):
        """Compile constraint to cvxpy.

        :param w_plus: Post-trade weights.
        :type w_plus: cvxpy.Variable
        :param z: Trade weights.
        :type z: cvxpy.Variable
        :param w_plus_minus_w_bm: Post-trade weights minus benchmark
            weights.
        :type w_plus_minus_w_bm: cvxpy.Variable
        :returns: Cvxpy constraints object.
        :rtype: cvxpy.constraints
        """
        return self._compile_constr_to_cvxpy(w_plus, z, w_plus_minus_w_bm) <=\
            self._rhs()

    def _compile_constr_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):
        """Cvxpy expression of the left-hand side of the constraint."""
        raise NotImplementedError # pragma: no cover

    def _rhs(self):
        """Cvxpy expression of the right-hand side of the constraint."""
        raise NotImplementedError # pragma: no cover


class CostInequalityConstraint(InequalityConstraint):
    """Linear inequality constraint applied to a cost term.

    The user does not interact with this class directly,
    it is returned by an expression such as ``cost <= value``
    where ``cost`` is a :class:`Cost` instance and ``value``
    is a scalar.
    """

    def __init__(self, cost, value):
        self.cost = cost
        self.value = DataEstimator(value, compile_parameter=True)

    def _compile_constr_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):
        """Compile constraint to cvxpy."""
        return self.cost.compile_to_cvxpy(w_plus, z, w_plus_minus_w_bm)

    def _rhs(self):
        return self.value.parameter

    def __repr__(self):
        return self.cost.__repr__() + ' <= ' + self.value.__repr__()



# Copyright 2016 Enzo Busseti, Stephen Boyd, Steven Diamond, BlackRock Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Unit tests for the cost objects."""

import unittest

import numpy as np
import pandas as pd

import cvxportfolio as cvx
from cvxportfolio.tests import CvxportfolioTest


class TestCosts(CvxportfolioTest):
    """Test base cost methods and tcost/hcost."""

    def test_cost_algebra(self):
        """Test algebraic operations on objective term instances."""
        # n = len(self.returns.columns)
        # wplus = cvx.Variable(n)
        self.w_plus_minus_w_bm.value = np.random.uniform(size=self.N)
        self.w_plus_minus_w_bm.value /= sum(self.w_plus_minus_w_bm.value)
        t = self.returns.index[1]

        cost1 = -.5 * cvx.DiagonalCovariance()
        cost2 = -.5 * cvx.FullCovariance(
            self.returns.iloc[:, :-1].T @ self.returns.iloc[:, :-1]
                / len(self.returns))
        cost3 = cost1 + cost2

        cost3.initialize_estimator_recursive(
            universe=self.returns.columns, trading_calendar=self.returns.index)
        expr3 = cost3.compile_to_cvxpy(
            self.w_plus, self.z, self.w_plus_minus_w_bm)
        expr1 = cost1.compile_to_cvxpy(
            self.w_plus, self.z, self.w_plus_minus_w_bm)
        expr2 = cost2.compile_to_cvxpy(
            self.w_plus, self.z, self.w_plus_minus_w_bm)
        cost3.values_in_time_recursive(
            t=t, past_returns=self.returns.loc[self.returns.index < t])
        self.assertTrue(expr3.value == expr1.value + expr2.value)

        cost4 = cost1 * 2
        expr4 = cost4.compile_to_cvxpy(
            self.w_plus, self.z, self.w_plus_minus_w_bm)
        self.assertTrue(expr4.value == expr1.value * 2)

        cost3 = cost1 + 3 * cost2
        expr3 = cost3.compile_to_cvxpy(
            self.w_plus, self.z, self.w_plus_minus_w_bm)
        self.assertTrue(expr3.value == expr1.value + 3 * expr2.value)

        cost3 = 3 * cost1 + 2 * cost2
        expr3 = cost3.compile_to_cvxpy(
            self.w_plus, self.z, self.w_plus_minus_w_bm)
        self.assertTrue(expr3.value == 3 * expr1.value + 2 * expr2.value)

        cost3 = .1 * cost1 + 2 * (cost2 + cost1)
        expr3 = cost3.compile_to_cvxpy(
            self.w_plus, self.z, self.w_plus_minus_w_bm)
        self.assertTrue(np.isclose(expr3.value, .1 * expr1.value +
                        2 * (expr2.value + expr1.value)))

        cost3 = cost1 + 5 * (cost2 + cost1)
        expr3 = cost3.compile_to_cvxpy(
            self.w_plus, self.z, self.w_plus_minus_w_bm)
        self.assertTrue(np.isclose(expr3.value, expr1.value +
                        5 * (expr2.value + expr1.value)))

    def test_extra_cost_algebra(self):
        """Test code of cost algebra that is not tested above."""

        cost = cvx.ReturnsForecast()
        with self.assertRaises(SyntaxError):
            _ = cost + 'hello'
        with self.assertRaises(SyntaxError): # __radd__
            _ = 'hello' + cost
        with self.assertRaises(SyntaxError):
            _ = cost - 2
        with self.assertRaises(TypeError): # __rsub__ , error thrown by - str
            _ = 'hello' - cost
        with self.assertRaises(SyntaxError):
            _ = 'hello' * cost
        with self.assertRaises(SyntaxError):
            _ = cost * 'hello'
        with self.assertRaises(TypeError):
            _ = cost <= 'hello'
        with self.assertRaises(TypeError):
            _ = cost >= 'hello'
        with self.assertRaises(SyntaxError):
            _ = cost < 0
        with self.assertRaises(SyntaxError):
            _ = cost > 0

    def test_hcost(self):
        """Test holding cost model."""
        dividends = pd.Series(np.random.randn(self.N-1),
                              self.returns.columns[:-1])

        for hcost, t in [
            [cvx.HoldingCost(short_fees=5, long_fees=3, dividends=dividends),
            100], # this is picked so that periods_per_year evaluates to 252
            [cvx.HoldingCost(
                periods_per_year=252, short_fees=5, long_fees=3,
                dividends=dividends),
            123],
        ]:

            hcost.initialize_estimator_recursive(
                universe=self.returns.columns,
                trading_calendar=self.returns.index)
            expression = hcost.compile_to_cvxpy(
                self.w_plus, self.z, self.w_plus_minus_w_bm)
            hcost.values_in_time_recursive(
                t=self.returns.index[t], past_returns=self.returns.iloc[:t])

            for _ in range(10):
                self.w_plus.value = np.random.randn(self.N)
                self.w_plus.value[-1] = 1 - np.sum(self.w_plus.value[:-1])

                print(expression.value)

                self.assertTrue(
                    np.isclose(expression.value,
                        # short fees
                        - np.sum(np.minimum(
                        self.w_plus.value[:-1], 0.))
                        * (np.exp(np.log(1.05)/252) - 1)
                        # long fees
                        + np.sum(np.maximum(
                        self.w_plus.value[:-1], 0.))
                        * (np.exp(np.log(1.03)/252) - 1)
                        # dividends
                        - self.w_plus.value[:-1].T @ dividends))

    def test_softconstraint(self):
        """Test code of SoftConstraints that is not covered elsewhere."""

        bad_constraint = cvx.NoTrade(
            asset='AAPL', periods=[pd.Timestamp.today()])
        bad_soft_constraint = cvx.SoftConstraint(bad_constraint)
        bad_soft_constraint.initialize_estimator_recursive(
            universe=['AAPL'], trading_calendar=[pd.Timestamp.today()])

        with self.assertRaises(SyntaxError):
            bad_soft_constraint.compile_to_cvxpy(
                self.w_plus, self.z, self.w_plus_minus_w_bm)

    def test_tcost(self):
        """Test tcost model."""
        value = 1e6

        pershare_cost = pd.Series(
            [0., 0.005, 0.],
            [self.returns.index[12], self.returns.index[23],
                self.returns.index[34]])
        b = pd.Series([0., 0., 1.], [self.returns.index[12],
                      self.returns.index[23], self.returns.index[34]])

        tcost = cvx.StocksTransactionCost(
            a=0.001/2, pershare_cost=pershare_cost,
            b=b, window_sigma_est=250, window_volume_est=250, exponent=1.5)

        tcost.initialize_estimator_recursive(
            universe=self.returns.columns, trading_calendar=self.returns.index)
        expression = tcost.compile_to_cvxpy(
            self.w_plus, self.z, self.w_plus_minus_w_bm)

        # only spread

        tcost.values_in_time_recursive(
            t=self.returns.index[12], current_portfolio_value=value,
            past_returns=self.returns.iloc[:12],
            past_volumes=self.volumes.iloc[:12],
            current_prices=pd.Series(np.ones(self.returns.shape[1]-1),
                self.returns.columns[:-1]))

        self.z.value = np.random.randn(self.returns.shape[1])
        self.z.value[-1] = -np.sum(self.z.value[:-1])

        est_tcost_lin = sum(np.abs(self.z.value[:-1]) * 0.0005)
        print(est_tcost_lin)
        print(expression.value)
        self.assertTrue(np.isclose(expression.value, est_tcost_lin))

        # spread and fixed cost

        prices = pd.Series(np.random.uniform(
            1, 100, size=self.returns.shape[1]-1), self.returns.columns[:-1])

        tcost.values_in_time_recursive(t=self.returns.index[23],
                                        current_portfolio_value=value,
                                        past_returns=self.returns.iloc[:23],
                                        past_volumes=self.volumes.iloc[:23],
                                        current_prices=prices)

        self.z.value = np.random.randn(self.returns.shape[1])
        self.z.value[-1] = -np.sum(self.z.value[:-1])

        est_tcost_lin = sum(np.abs(self.z.value[:-1]) * 0.0005)
        est_tcost_lin += np.abs(self.z.value[:-1]) @ (0.005 / prices)
        print(est_tcost_lin)
        print(expression.value)
        self.assertTrue(np.isclose(expression.value, est_tcost_lin))

        # spread and nonlin cost

        tcost.values_in_time_recursive(
            t=self.returns.index[34],
            current_portfolio_value=value,
            past_returns=self.returns.iloc[:34],
            past_volumes=self.volumes.iloc[:34],
            current_prices=pd.Series(np.ones(self.returns.shape[1]-1),
                self.returns.columns[:-1]))

        self.z.value = np.random.randn(self.returns.shape[1])
        self.z.value[-1] = -np.sum(self.z.value[:-1])

        est_tcost_lin = sum(np.abs(self.z.value[:-1]) * 0.0005)
        volumes_est = self.volumes.iloc[:34].mean().values
        sigmas_est = np.sqrt((self.returns.iloc[:34, :-1]**2).mean()).values
        est_tcost_nonnlin = (
            np.abs(self.z.value[:-1])**(3/2)) @ (
                sigmas_est * np.sqrt(value / volumes_est))
        print(est_tcost_lin)
        print(est_tcost_nonnlin)
        print(expression.value)
        self.assertTrue(np.isclose(expression.value,
                        est_tcost_lin+est_tcost_nonnlin))

        # also c
        tcost = cvx.StocksTransactionCost(
            a=0.001/2, pershare_cost=pershare_cost,
            b=b, window_sigma_est=250, window_volume_est=250, c=0.001,
             exponent=1.5)
        tcost.initialize_estimator_recursive(
            universe=self.returns.columns, trading_calendar=self.returns.index)
        expression = tcost.compile_to_cvxpy(
            self.w_plus, self.z, self.w_plus_minus_w_bm)

        tcost.values_in_time_recursive(
            t=self.returns.index[34],
            current_portfolio_value=value,
            past_returns=self.returns.iloc[:34],
            past_volumes=self.volumes.iloc[:34],
            current_prices=pd.Series(np.ones(self.returns.shape[1]-1),
                self.returns.columns[:-1]))

        self.z.value = np.random.randn(self.returns.shape[1])
        self.z.value[-1] = -np.sum(self.z.value[:-1])

        est_tcost_lin = sum(np.abs(self.z.value[:-1]) * 0.0005)
        volumes_est = self.volumes.iloc[:34].mean().values
        sigmas_est = np.sqrt((self.returns.iloc[:34, :-1]**2).mean()).values
        est_tcost_nonnlin = (
            np.abs(self.z.value[:-1])**(3/2)) @ (
                sigmas_est * np.sqrt(value / volumes_est))
        est_tcost_c = np.sum(self.z.value[:-1] * 0.001)
        print(est_tcost_lin)
        print(est_tcost_nonnlin)
        print(expression.value)
        self.assertTrue(np.isclose(expression.value,
                        est_tcost_lin+est_tcost_nonnlin+est_tcost_c))


if __name__ == '__main__':

    unittest.main() # pragma: no cover



# Copyright 2016 Enzo Busseti, Stephen Boyd, Steven Diamond, BlackRock Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Unit tests for the return forecast objects."""

import unittest

import numpy as np

from cvxportfolio.returns import (CashReturn, ReturnsForecast,
                                  ReturnsForecastError)
from cvxportfolio.tests import CvxportfolioTest


class TestReturns(CvxportfolioTest):
    """Test return objects."""

    def test_cash_returns(self):
        """Test CashReturn object with last return as forecast."""
        cash_model = CashReturn()
        cvxpy_expression = self.boilerplate(cash_model)
        self.w_plus.value = np.random.randn(self.N)
        cash_model.values_in_time_recursive(
            t=None, past_returns=self.returns.iloc[:123])
        cr = self.returns.iloc[122, -1]
        self.assertTrue(cvxpy_expression.value == cr * (
            self.w_plus[-1].value ))

    def test_cash_returns_provided(self):
        """Test CashReturn object with provided cash returns."""
        cash_model = CashReturn(self.returns.iloc[:, -1])
        cvxpy_expression = self.boilerplate(cash_model)
        self.w_plus.value = np.random.randn(self.N)
        cash_model.values_in_time_recursive(
            t=self.returns.index[123], past_returns=None)
        cr = self.returns.iloc[123, -1]
        self.assertTrue(cvxpy_expression.value == cr * (
            self.w_plus[-1].value))

    def test_returns_forecast(self):
        """Test ReturnsForecast object with provided assets' returns."""
        alpha_model = ReturnsForecast(self.returns.iloc[:, :-1])
        cvxpy_expression = self.boilerplate(alpha_model)
        alpha_model.values_in_time_recursive(
            t=self.returns.index[123], past_returns=None)
        self.w_plus.value = np.random.randn(self.N)
        print(cvxpy_expression.value)
        print(self.w_plus[:-1].value @ self.returns.iloc[123][:-1])
        self.assertTrue(np.isclose(cvxpy_expression.value,
                        self.w_plus[:-1].value @ self.returns.iloc[123][:-1]))

    def test_full_returns_forecast(self):
        """Test ReturnsForecast object with historical mean forecasts."""
        alpha_model = ReturnsForecast()
        cvxpy_expression = self.boilerplate(alpha_model)
        t = self.returns.index[123]
        alpha_model.values_in_time_recursive(
            t=t, past_returns=self.returns.loc[self.returns.index < t])
        self.w_plus.value = np.random.uniform(size=self.N)
        self.w_plus.value /= sum(self.w_plus.value)
        myforecast = self.returns.iloc[:, :-
                                       1].loc[self.returns.index < t].mean()
        self.assertTrue(np.isclose(cvxpy_expression.value,
                        self.w_plus.value[:-1] @ myforecast))

    def test_returns_forecast_error(self):
        """Test ReturnsForecastError object with provided values."""
        delta = self.returns.iloc[:, :-
                                  1].std(ddof=0) / np.sqrt(len(self.returns))
        error_risk = ReturnsForecastError(delta)
        cvxpy_expression = self.boilerplate(error_risk)
        error_risk.values_in_time_recursive(t='ciao', past_returns='hello')
        self.w_plus_minus_w_bm.value = np.random.randn(self.N)
        self.assertTrue(np.isclose(cvxpy_expression.value, np.abs(
            self.w_plus_minus_w_bm.value[:-1]) @ delta))

    def test_full_returns_forecast_error(self):
        """Test ReturnsForecastError object with as forecast the std of the.

        mean estimator.
        """
        error_risk = ReturnsForecastError()
        cvxpy_expression = self.boilerplate(error_risk)
        t = self.returns.index[123]
        past_returns = self.returns.loc[self.returns.index < t]
        error_risk.values_in_time_recursive(t=t, past_returns=past_returns)
        self.w_plus_minus_w_bm.value = np.random.randn(self.N)
        delta = past_returns.std(ddof=0) / np.sqrt(past_returns.count())
        print(cvxpy_expression.value)
        print(np.abs(self.w_plus_minus_w_bm.value[:-1]) @ delta[:-1])
        self.assertTrue(np.isclose(cvxpy_expression.value, np.abs(
            self.w_plus_minus_w_bm.value[:-1]) @ delta[:-1]))


if __name__ == '__main__':

    unittest.main() # pragma: no cover



# Copyright 2023 Enzo Busseti
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module implements Estimator base classes.

Policies, costs, and constraints inherit from this.
"""

import numbers

import cvxpy as cp
import numpy as np
import pandas as pd

from .errors import DataError, MissingAssetsError, MissingTimesError, NaNError
from .utils import make_numeric, repr_numpy_pandas


class Estimator:
    """Estimator abstraction, designed for repeated evaluation over time.

    Policies, costs, and constraints inherit from this. When overloading
    methods defined here one should be very careful. The recommended usage
    (if you want a class that uses our recursive execution model) is to
    put any sub-estimators at the class attribute level, like we do
    throughout the library. That ensures that the sub-estimators will
    be evaluated before the class itself by both
    :meth:`initialize_estimator_recursive` and
    :meth:`values_in_time_recursive`.
    """

    def initialize_estimator(self, universe, trading_calendar, **kwargs):
        """Initialize estimator instance with universe and trading times.

        This method is called at the start of an online execution, or, in a
        back-test, at its start and whenever the trading universe changes.
        It provides the instance with the current trading universe and a
        :class:`pandas.DatetimeIndex` representing the current and future
        trading calendar, *i.e.*, the times at which the estimator will be
        evaluated, or a best guess of it. The instance uses these to
        appropriately initialize any internal object, such as Cvxpy parameters,
        to the right size (as implied by the universe). Also, especially for
        multi-period optimization and similar policies, awareness of the future
        trading calendar is essential to, *e.g.*, plan in advance.

        :param universe: Trading universe, including cash.
        :type universe: pandas.Index
        :param trading_calendar: Future (including current) trading calendar.
        :type trading_calendar: pandas.DatetimeIndex
        :param kwargs: Reserved for future expansion.
        :type kwargs: dict
        """

        # we don't raise NotImplementedError because this is called
        # on classes that don't re-define it

    def initialize_estimator_recursive(self, **kwargs):
        """Recursively initialize all estimators in a policy.

        :param kwargs: Parameters sent down an estimator tree to inizialize it.
        :type kwargs: dict
        """
        for _, subestimator in self.__dict__.items():
            if hasattr(subestimator, "initialize_estimator_recursive"):
                subestimator.initialize_estimator_recursive(**kwargs)
        if hasattr(self, "initialize_estimator"):
            self.initialize_estimator(**kwargs)

    def finalize_estimator(self, **kwargs):
        """Finalize estimator instance (currently unused).

        This method is called at the end of an online execution, or, in a
        back-test, whenever the trading universe changes (before calling
        :meth:`initialize_estimator` with the new universe) and at its end.
        We aren't currently using in the rest of the library but we plan to
        move the caching logic in it.

        .. versionadded:: 1.1.0

        :param kwargs: Reserved for future expansion.
        :type kwargs: dict
        """

        # we don't raise NotImplementedError because this is called
        # on classes that don't re-define it

    def finalize_estimator_recursive(self, **kwargs):
        """Recursively finalize all estimators in a policy.

        .. versionadded:: 1.1.0

        :param kwargs: Parameters sent down an estimator tree to finalize it.
        :type kwargs: dict
        """
        for _, subestimator in self.__dict__.items():
            if hasattr(subestimator, "finalize_estimator_recursive"):
                subestimator.finalize_estimator_recursive(**kwargs)
        if hasattr(self, "finalize_estimator"):
            self.finalize_estimator(**kwargs)

    _current_value = None

    @property
    def current_value(self):
        """Current value of this instance.

        :returns: Current value, which can be any object.
        :rtype: numpy.array, pandas.Series, pandas.DataFrame, ...
        """
        return self._current_value

    # pylint: disable=too-many-arguments
    def values_in_time(
            self, t, current_weights, current_portfolio_value,
            past_returns, past_volumes, current_prices,
            mpo_step=None, cache=None, **kwargs):
        """Evaluate estimator at current time, possibly return current value.

        This method is usually the most important for Estimator classes.
        It is called at each point in a back-test with all data of the current
        state. Sub-estimators are evaluated first, in a depth-first recursive
        tree fashion (defined in :meth:`values_in_time_recursive`). The
        signature differs slightly between different estimators, see below.

        :param t: Current timestamp.
        :type t: pandas.Timestamp
        :param current_weights: Current allocation weights.
        :type current_weights: pandas.Series
        :param current_portfolio_value: Current total value of the portfolio
            in cash units.
        :type current_portfolio_value: float
        :param past_returns: Past market returns (including cash).
        :type past_returns: pandas.DataFrame
        :param past_volumes: Past market volumes, or None if not available.
        :type past_volumes: pandas.DataFrame or None
        :param current_prices: Current (open) prices, or None if not available.
        :type current_prices: pandas.Series or None
        :param mpo_step: For :class:`cvxportfolio.MultiPeriodOptimization`
            which step in future planning this estimator is at: 0 is for
            the current step (:class:`cvxportfolio.SinglePeriodOptimization`),
            1 is for day ahead, .... Defaults to ``None`` if unused.
        :type mpo_step: int, optional
        :param cache: Cache or workspace shared between all elements of an
            estimator tree, currently only used by
            :class:`cvxportfolio.MultiPeriodOptimization` (and derived
            classes). It's useful to avoid re-computing expensive things like
            covariance estimates at different MPO steps. Defaults to ``None``
            if unused.
        :type cache: dict, optional
        :param kwargs: Reserved for future expansion.
        :type kwargs: dict

        :returns: Current value of the estimator.
        :rtype: object or None
        """
        # we don't raise NotImplementedError because this is called
        # on classes that don't re-define it

    def values_in_time_recursive(self, **kwargs):
        """Evaluate recursively on sub-estimators.

        :param kwargs: All parameters to :meth:`values_in_time` that are passed
            to all elements contained in a policy object.
        :type kwargs: dict

        :returns: The current value evaluated by this instance, if it
            implements the :meth:`values_in_time` method and it returns
            something there.
        :rtype: numpy.array, pandas.Series, pandas.DataFrame, ...
        """
        for _, subestimator in self.__dict__.items():
            if hasattr(subestimator, "values_in_time_recursive"):
                subestimator.values_in_time_recursive(**kwargs)
        if hasattr(self, "values_in_time"):
            # pylint: disable=assignment-from-no-return
            self._current_value = self.values_in_time(**kwargs)
            return self.current_value
        return None

    def collect_hyperparameters(self):
        """Collect (recursively) all hyperparameters defined in a policy.

        :returns: List of :class:`cvxportfolio.hyperparameters.HyperParameter`
            instances.
        :rtype: list
        """
        result = []
        for _, subestimator in self.__dict__.items():
            if hasattr(subestimator, "collect_hyperparameters"):
                result += subestimator.collect_hyperparameters()
        return result

    def __repr__(self):
        """Pretty-print the cvxportfolio object in question.

        We make sure that every object the user interacts with can be
        pretty-printed to the interpreter, ideally in a way such that
        copy-pasting the output to the prompt results in an identical object.
        We can't do that all the times but we do our best. This is used
        throughout the library, for example it is included in backtest results
        so the user knows which policy generated that backtest, .... We prefer
        to define the logic of this directly insted of relying, e.g., on
        dataclasses logic, because we want to customize it to our usecase.
        """
        lhs = self.__class__.__name__ + '('
        core = ''
        for name, attr in self.__dict__.items():
            if attr is None:
                continue
            if hasattr(attr, "values_in_time_recursive") or\
                    hasattr(attr, "values_in_time") or (name[0] != '_'):
                core += name + '=' + attr.__repr__() + ', '
        core = core[:-2]  # remove trailing comma and space if present
        rhs = ')'
        return lhs + core + rhs


class SimulatorEstimator(Estimator):
    """Base class for estimators that are used by the market simulator.

    .. versionadded:: 1.2.0

    This is currently used as the base class for
    :class:`cvxportfolio.costs.SimulatorCost` but could implement in future
    versions more operations done as part of a simulation (back-test) loop,
    like filtering out trades (rejecting small ones, ...), or more. It allows
    for nested evaluation like it's done by :meth:`values_in_time` and
    :meth:`values_in_time_recursive`. Estimators that are used in the market
    simulator should derive from this. Some examples are
    :class:`DataEstimator`, which is used to select values (like borrow costs)
    in simulations as well as in optimization, and
    :class:`cvxportfolio.forecast.HistoricalStandardDeviation`, which is used
    also in simulation by :class:`cvxportfolio.costs.TransactionCost`.
    """

    def simulate( # pylint: disable=too-many-arguments
        self, t, t_next, u, h_plus, past_volumes,
        past_returns, current_prices, current_returns, current_volumes,
        current_weights, current_portfolio_value, **kwargs):
        """Evaluate the estimator as part of a Market Simulator back-test loop.

        Cost classes that are meant to be used in the simulator can
        implement this. The arguments to this are the same as for
        :meth:`cvxportfolio.estimator.Estimator.values_in_time` plus the
        realized returns and volumes in the period, and the trades requested
        by the policy, ....

        :param t: Current timestamp.
        :type t: pandas.Timestamp
        :param u: Trade vector in cash units requested by the policy.
            If the market simulator implements rounding by number of shares
            and/or canceling trades on assets whose volume for the period
            is zero, this is after those transformations.
        :type u: pandas.Series
        :param h_plus: Post-trade holdings vector.
        :type h_plus: pandas.Series
        :param past_returns: Past market returns (including cash).
        :type past_returns: pandas.DataFrame
        :param current_returns: Current period's market returns (including
            cash).
        :type current_returns: pandas.Series
        :param past_volumes: Past market volumes, or None if not available.
        :type past_volumes: pandas.DataFrame or None
        :param current_volumes: Current period's market volumes, or None if not
            available.
        :type current_volumes: pandas.Series or None
        :param current_prices: Current (open) prices, or None if not available.
        :type current_prices: pandas.Series or None
        :param current_weights: Current allocation weights (before trading).
        :type current_weights: pandas.Series
        :param current_portfolio_value: Current total value of the portfolio
            in cash units, before costs.
        :type current_portfolio_value: float
        :param t_next: Timestamp of the next trading period.
        :type t_next: pandas.Timestamp
        :param kwargs: Reserved for future expansion.
        :type kwargs: dict

        :returns: The current value of this instance.
        :rtype: any object
        """
        raise NotImplementedError # pragma: no cover

    def simulate_recursive(self, **kwargs):
        """Evaluate simulator value(s) recursively on sub-estimators.

        :param kwargs: All parameters to :meth:`simulate` that are passed
            to all elements contained in a simulator cost object.
        :type kwargs: dict

        :returns: The current simulator value evaluated by this instance, if it
            implements the :meth:`simulate` method and it returns
            something there.
        :rtype: numpy.array, pandas.Series, pandas.DataFrame, ...
        """
        for _, subestimator in self.__dict__.items():
            if hasattr(subestimator, "simulate_recursive"):
                subestimator.simulate_recursive(**kwargs)
        if hasattr(self, "simulate"):
            self._current_value = self.simulate(**kwargs)
            return self.current_value
        return None # pragma: no cover

class CvxpyExpressionEstimator(Estimator):
    """Base class for estimators that are Cvxpy expressions."""

    def compile_to_cvxpy(self, w_plus, z, w_plus_minus_w_bm):
        """Compile term to cvxpy expression.

        This is called by a Policy class on its terms before the start
        of the backtest to compile its Cvxpy problem. If the Policy
        changes in time this is called at every time step.

        It can either return a scalar expression, in the case of
        objective terms, or a list of cvxpy constraints, in the case of
        constraints.

        In MultiPeriodOptimization policies this is called separately
        for costs and constraints at different look-ahead steps with the
        corresponding w_plus and z.

        :param w_plus: Post-trade weights.
        :type w_plus: cvxpy.Variable
        :param z: Trade weights.
        :type z: cvxpy.Variable
        :param w_plus_minus_w_bm: Post-trade weights minus benchmark
            weights.
        :type w_plus_minus_w_bm: cvxpy.Variable
        """
        raise NotImplementedError

# pylint: disable=too-many-arguments
class DataEstimator(SimulatorEstimator):
    """Estimator of point-in-time values from internal data.

    It also implements logic to check that no ``nan`` are returned
    by its ``values_in_time_recursive`` method, which is the way Cvxportfolio
    objects use this class to get data, to compile and update a Cvxpy
    parameter, and to slice the data with the current trading universe.

    :param data: Data expressed preferably as pandas Series or DataFrame
        where the first index is a ``pandas.DateTimeIndex``. Otherwise you can
        pass a callable object which implements the
        :meth:`values_in_time_recursive` method (with the standard signature)
        and returns the corresponding value in time, or a constant float,
        numpy.array, or even pandas Series or DataFrame not indexed by time
        (e.g., a covariance matrix where both index and columns are the stock
        symbols).
    :type data: object, pandas.Series, pandas.DataFrame
    :param use_last_available_time: if the pandas index exists
        and is a ``pandas.DateTimeIndex`` you can instruct
        :meth:`values_in_time_recursive` to retrieve the last available value
        at time t by setting this to True. Default is False.
    :type use_last_available_time: bool
    :param allow_nans: If True, allow data returned to contain ``nan``.
        Default False.
    :type allow_nans: bool
    :param compile_parameter: If True, compile a Cvxpy parameter that gets
        updated with the current value of the instance at each point in a
        backtest. Default False.
    :type compile_parameter: bool
    :param non_negative: If True, the compiled Cvxpy parameter is non-negative
        (this affects certain Cvxpy operations). Default False.
    :type non_negative: bool
    :param positive_semi_definite: If True, the compiled Cvxpy parameter is
        market as a positive semi-definite matrix (this affects certain Cvxpy
        operations). Default False.
    :type positive_semi_definite: bool
    :param data_includes_cash: If True, when the data is sliced with the
        current trading universe we also look for the values corresponding to
        the cash account. Default False.
    :type data_includes_cash: bool
    :param ignore_shape_check: If True, we don't do any slicing of the data
        according to the current trading universe. Default False.
    :type ignore_shape_check: bool

    :raises cvxportfolio.NaNError: If ``nan`` are present in result.
    :raises cvxportfolio.MissingTimesError: If some times are missing.
    :raises cvxportfolio.MissingAssetsError: If some assets are missing.
    :raises cvxportfolio.DataError: If data is not in the right form.
    """

    def __init__(
            self, data, use_last_available_time=False, allow_nans=False,
            compile_parameter=False, non_negative=False,
            positive_semi_definite=False, data_includes_cash=False,
            ignore_shape_check=False):
        self.data = make_numeric(data)
        self._use_last_available_time = use_last_available_time
        self._allow_nans = allow_nans
        self._compile_parameter = compile_parameter
        self._non_negative = non_negative
        self._positive_semi_definite = positive_semi_definite
        self._universe_maybe_noncash = None
        self._data_includes_cash = data_includes_cash
        self._ignore_shape_check = ignore_shape_check
        self.parameter = None

    def initialize_estimator(self, universe, trading_calendar, **kwargs):
        """Initialize with current universe.

        :param universe: Trading universe, including cash.
        :type universe: pandas.Index
        :param trading_calendar: Future (including current) trading calendar.
        :type trading_calendar: pandas.DatetimeIndex
        :param kwargs: Other unused arguments to :meth:`initialize_estimator`.
        :type kwargs: dict
        """

        self._universe_maybe_noncash = \
            universe if self._data_includes_cash else universe[:-1]

        if self._compile_parameter:
            value = self._internal_values_in_time(
                t=trading_calendar[0])
            self.parameter = cp.Parameter(
                value.shape if hasattr(value, "shape") else (),
                PSD=self._positive_semi_definite, nonneg=self._non_negative)

    def value_checker(self, result):
        """Ensure that only scalars or arrays without np.nan are returned.

        :raises cvxportfolio.errors.NaNError: If NaNs are found.
        :raises cvxportfolio.errors.DataError: If the value passed is not a
            scalar or Numpy array.

        :param result: Scalar or array that we check has no NaNs.
        :type result: float or numpy.array

        :returns: Input value; if array, a copy.
        :rtype: float or numpy.array
        """

        if isinstance(result, numbers.Number):
            if np.isnan(result) and not self._allow_nans:
                raise NaNError(
                    f"{self.__class__.__name__}.values_in_time_recursive"
                    + " result is a np.nan scalar.")
            return result

        if isinstance(result, np.ndarray):
            if np.any(np.isnan(result)) and not self._allow_nans:
                message = f"{self.__class__.__name__}.values_in_time_recursive"
                message += " result is an array with np.nan's."
                raise NaNError(message)
            # we pass a copy because it can be accidentally overwritten
            return np.array(result)

        raise DataError(
            f"{self.__class__.__name__}.values_in_time_recursive result"
            + " is not a scalar or array.")

    def _universe_subselect(self, data):
        """This function subselects from ``data`` the relevant universe.

        See github issue #106.

        If data is a pandas Series we subselect its index. If we fail we
        throw an error. If data is a pandas DataFrame (covariance,
        exposure matrix) we try to subselect its index and columns. If
        we fail on either we ignore the failure, but if we fail on both
        we throw an error. If data is a numpy 1-d array we check that
        its length is the same as the universe's. If it is a 2-d array
        we check that at least one dimension is the same as the
        universe's. If the universe is None we skip all checks. (We may
        revisit this choice.) This only happens if the DataEstimator
        instance is not part of a Estimator tree (a usecase which we
        will probably drop).
        """

        if (self._universe_maybe_noncash is None) or self._ignore_shape_check:
            return data.values if hasattr(data, 'values') else data

        if isinstance(data, pd.Series):
            try:
                return data.loc[self._universe_maybe_noncash].values
            except KeyError as exc:
                raise MissingAssetsError(
                    "The pandas Series found by %s has index %s"
                    + " while the current universe%s"
                    + " is %s. It was not possible to reconcile the two.",
                    self.__class__.__name__, data.index,
                    ' minus cash' if not self._data_includes_cash else ' ',
                    self._universe_maybe_noncash) from exc

        if isinstance(data, pd.DataFrame):
            try:
                return data.loc[self._universe_maybe_noncash,
                    self._universe_maybe_noncash].values
            except KeyError:
                try:
                    return data.loc[:, self._universe_maybe_noncash].values
                except KeyError:
                    try:
                        return data.loc[self._universe_maybe_noncash, :].values
                    except KeyError:
                        pass
            raise MissingAssetsError(
                "The pandas DataFrame found by %s has index %s"
                + " and columns %s"
                + " while the current universe%s"
                + " is %s. It was not possible to reconcile the two.",
                self.__class__.__name__, data.columns,
                ' minus cash' if not self._data_includes_cash else ' ',
                self._universe_maybe_noncash)

        if isinstance(data, np.ndarray):
            dimensions = data.shape
            if not len(self._universe_maybe_noncash) in dimensions:
                raise MissingAssetsError(
                    "The numpy array found by %s has dimensions %s"
                    + " while the current universe%s "
                    + "has size %s. It was not possible to reconcile the two.",
                    self.__class__.__name__, data.shape,
                    ' minus cash' if not self._data_includes_cash else ' ',
                    len(self._universe_maybe_noncash))
            return data

        # scalar
        return data

    def _internal_values_in_time(self, t, **kwargs):
        """Internal method called by :meth:`values_in_time`."""

        # here we trust the result (change?)
        if hasattr(self.data, "values_in_time_recursive"):
            return self.data.current_value

        # here (probably user-provided) we check
        if hasattr(self.data, "values_in_time"):
            if len(kwargs) == 0:
                raise ValueError(
                    "It seems you're using a custom forecaster as part of a "
                    "simulate_recursive evaluation, you should derive from "
                    "SimulatorEstimator instead.")
            return self.value_checker(self._universe_subselect(
                self.data.current_value if hasattr(self.data, 'current_value')
                else self.data.values_in_time(t=t, **kwargs)))

        # if self.data is pandas and has datetime (first) index
        if (hasattr(self.data, "loc") and hasattr(self.data, "index")
            and (isinstance(self.data.index, pd.DatetimeIndex)
                 or (isinstance(self.data.index, pd.MultiIndex) and
                     isinstance(self.data.index.levels[0],
                         pd.DatetimeIndex)))):
            try:
                if self._use_last_available_time:
                    if isinstance(self.data.index, pd.MultiIndex):
                        newt = self.data.index.levels[0][
                            self.data.index.levels[0] <= t][-1]
                    else:
                        newt = self.data.index[self.data.index <= t][-1]
                    tmp = self.data.loc[newt]
                else:
                    tmp = self.data.loc[t]

                return self.value_checker(self._universe_subselect(tmp))

            except (KeyError, IndexError) as exc:
                raise MissingTimesError(
                    "%s.values_in_time_recursive could not find data"
                    + " for time %s. This could be due to wrong timezone"
                    + " setting: in general Cvxportfolio objects are timezone"
                    + " aware, the data you pass should be as well.",
                     self, t) from exc

        # if data is pandas but no datetime index (constant in time)
        if hasattr(self.data, "values"):
            return self.value_checker(self._universe_subselect(self.data))

        # if data is scalar or numpy
        return self.value_checker(self._universe_subselect(self.data))

    def values_in_time( # pylint: disable=arguments-differ
            self, **kwargs):
        """Obtain value of `self.data` at time t or right before.

        :param kwargs: All parameters passed to :meth:`values_in_time`.
        :type kwargs: dict

        :raises cvxportfolio.errors.NaNError: The data provided contains
            NaNs at current time.

        :returns: The  value from this
            :class:`cvxportfolio.estimator.DataEstimator` at current time.
        :rtype: int, float, numpy.ndarray
        """
        try:
            result = self._internal_values_in_time(**kwargs)
        except NaNError as exc:
            raise NaNError(f"{self.__class__.__name__} found NaNs"
                + f" at time {kwargs['t']}.") from exc
        if self.parameter is not None:
            self.parameter.value = result
        return result

    def simulate( # pylint: disable=arguments-differ
            self, t, **kwargs):
        """Evaluate in simulation (e.g., TransactionCost).

        :param t: Current timestamp.
        :type t: pd.Timestamp
        :param kwargs: All other unused arguments to
            :meth:`SimulatorEstimator.simulate`.
        :type kwargs: dict

        :returns: The  value from this
            :class:`cvxportfolio.estimator.DataEstimator` at current time.
        :rtype: int, float, numpy.ndarray
        """
        # We don't support evaluation inside DataEstimator in this case
        # You should implement simulate/simulate_recursive in each
        # SimulatorEstimator wrapped by DataEstimator
        return self.values_in_time(t=t)

    def __repr__(self):
        """Pretty-print."""
        if np.isscalar(self.data):
            return str(self.data)
        if hasattr(self.data, 'values_in_time_recursive'
            ) or hasattr(self.data, 'values_in_time'):
            return self.data.__repr__()
        return repr_numpy_pandas(self.data)



# Copyright 2016 Enzo Busseti, Stephen Boyd, Steven Diamond, BlackRock Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Unit tests for the data and parameter estimator objects."""

import unittest

import numpy as np
import pandas as pd

from cvxportfolio.forecast import (ForecastError, HistoricalCovariance,
                                   HistoricalFactorizedCovariance,
                                   HistoricalLowRankCovarianceSVD,
                                   HistoricalMeanError, HistoricalMeanReturn,
                                   HistoricalMeanVolume,
                                   HistoricalStandardDeviation,
                                   HistoricalVariance)
from cvxportfolio.tests import CvxportfolioTest


class TestForecast(CvxportfolioTest):
    """Test forecast estimators and their caching.

    In most cases we test against the relevant pandas function as reference.
    """

    def test_historical_mean_volume(self):
        """Test mean volume forecaster."""

        forecaster = HistoricalMeanVolume()
        for tidx in [20, 21, 22, 25, 26, 27]:
            cvx_val = forecaster.values_in_time_recursive(
                t=self.volumes.index[tidx],
                past_returns=self.returns.iloc[:tidx],
                past_volumes=self.volumes.iloc[:tidx])

            self.assertTrue(np.allclose(cvx_val,
                self.volumes.iloc[:tidx].mean()))

        with self.assertRaises(ValueError):
            cvx_val = forecaster.values_in_time_recursive(
                t=self.volumes.index[tidx],
                past_returns=self.returns.iloc[:tidx],
                past_volumes=None)

    def test_vector_fc_syntax(self):
        """Test syntax of vector forecasters."""
        with self.assertRaises(ValueError):
            forecaster = HistoricalMeanReturn(rolling=3)
            forecaster.values_in_time_recursive(
                t=pd.Timestamp.today(), past_returns=self.returns)

        with self.assertRaises(ValueError):
            forecaster = HistoricalMeanReturn(rolling=pd.Timedelta('0d'))
            forecaster.values_in_time_recursive(
                t=pd.Timestamp.today(), past_returns=self.returns)

        with self.assertRaises(ForecastError):
            forecaster = HistoricalMeanReturn()
            returns = pd.DataFrame(self.returns, copy=True)
            returns.iloc[:40, 3:10] = np.nan
            forecaster.values_in_time_recursive(
                t=self.returns.index[20], past_returns=returns.iloc[:20])

        # test that update throws exception when moving window results
        # in invalid data
        forecaster = HistoricalMeanReturn(rolling=pd.Timedelta('10d'))
        returns = pd.DataFrame(self.returns, copy=True)
        returns.iloc[20:, 3:10] = np.nan
        last_valid_t = 25
        forecaster.values_in_time_recursive(
            t=self.returns.index[last_valid_t],
            past_returns=returns.iloc[:last_valid_t])

        with self.assertRaises(ForecastError):
            forecaster.values_in_time_recursive(
                t=self.returns.index[last_valid_t+1],
                past_returns=returns.iloc[:last_valid_t+1])

    def _base_test_vector_update(
            # pylint: disable=too-many-arguments
            self, forecaster, fc_kwargs, df_attr, pd_kwargs, df_callable=None,
            with_nans=True):
        """Base function to test vector updates (mean, std, var)."""
        forecaster = forecaster(**fc_kwargs)

        returns = pd.DataFrame(self.returns, copy=True)
        if with_nans:
            returns.iloc[:40, 3:10] = np.nan

        for tidx in [50, 51, 52, 55, 56, 57]:
            t = returns.index[tidx]
            past_returns = returns.loc[returns.index < t]
            cvx_val = forecaster.values_in_time_recursive(
                t=t, past_returns=past_returns)
            self.assertFalse(np.any(np.isnan(cvx_val)))
            self.assertTrue(np.allclose(
                cvx_val,
                df_callable(past_returns.iloc[:, :-1])
                    if df_callable is not None else
                    getattr(past_returns.iloc[:, :-1], df_attr)(**pd_kwargs)))

    def test_mean_update(self):
        """Test the mean forecaster."""
        self._base_test_vector_update(HistoricalMeanReturn, {}, 'mean', {})

    def test_variance_update(self):
        """Test the variance forecaster."""
        self._base_test_vector_update(
            HistoricalVariance, {'kelly': False}, 'var', {'ddof': 0})

    def test_stddev_update(self):
        """Test the standard deviation forecaster."""
        self._base_test_vector_update(
            HistoricalStandardDeviation, {'kelly': False}, 'std', {'ddof': 0})

    def test_meanerror_update(self):
        """Test the forecaster of the standard deviation on the mean."""

        def _me(past_returns_noncash):
            return past_returns_noncash.std(ddof=0)\
                / np.sqrt(past_returns_noncash.count())

        self._base_test_vector_update(
            HistoricalMeanError, {}, None, None, df_callable=_me)

    def _kelly_covariance(self, past_returns_noncash, half_life=None):
        """Covariance using Kelly.

        Below we implement also the one not using Kelly. That should be
        done by past_returns_noncash.cov(ddof=0) but a bug in pandas
        prevents from doing so (see below).
        """

        # EMW
        if half_life is not None:
            index_in_halflifes = (
                past_returns_noncash.index - past_returns_noncash.index[-1]
                    ) / half_life
            emw_weights = np.exp(index_in_halflifes * np.log(2))
        else:
            emw_weights = pd.Series(1., past_returns_noncash.index)

        _ = past_returns_noncash.fillna(0.)
        num = _.T @ _.multiply(emw_weights, axis=0)
        _ = (~past_returns_noncash.isnull()) * 1.
        den = _.T @ _.multiply(emw_weights, axis=0)

        return num / den

    def _nokelly_covariance_emw_nonans(self, past_returns_noncash, half_life):
        """This is only without nans."""
        result = self._kelly_covariance(
            past_returns_noncash, half_life=half_life)
        means = self._mean_emw(past_returns_noncash, half_life)
        return result - np.outer(means, means)

    def test_cov_update(self):
        """Test the covariance forecaster."""

        self._base_test_vector_update(
            HistoricalCovariance, {'kelly': True}, None, None,
            df_callable=self._kelly_covariance)

    def test_cov_update_nokelly(self):
        """Test the covariance forecaster without Kelly correction.

        Due to a bug in pandas, we can compare with Pandas' DataFrame.cov
        only if the df has no NaNs.
        """

        self._base_test_vector_update(
            HistoricalCovariance, {'kelly': False}, 'cov', {'ddof': 0},
            with_nans=False)

    def _base_test_moving_window_vector_update(
            # pylint: disable=too-many-arguments
            self, forecaster, fc_kwargs, df_attr, pd_kwargs, df_callable=None,
            with_nans=True):
        """Base test for vector quantities using a moving window."""

        window = pd.Timedelta('20d')
        forecaster = forecaster(**fc_kwargs, rolling=window)

        returns = pd.DataFrame(self.returns, copy=True)
        if with_nans:
            returns.iloc[:40, 3:10] = np.nan

        for tidx in [50, 51, 52, 55, 56, 57]:
            t = returns.index[tidx]
            past_returns = returns.loc[returns.index < t]
            past_returns_window = past_returns.loc[
                past_returns.index >= t - window]
            cvx_val = forecaster.values_in_time_recursive(
                t=t, past_returns=past_returns)

            self.assertFalse(np.any(np.isnan(cvx_val)))

            test_val = (df_callable(past_returns_window.iloc[:, :-1])
                    if df_callable is not None else
                    getattr(past_returns_window.iloc[:, :-1],
                        df_attr)(**pd_kwargs))

            self.assertTrue(np.allclose(cvx_val, test_val))

    def test_mean_update_moving_window(self):
        """Test the mean forecaster with moving window."""
        self._base_test_moving_window_vector_update(
            HistoricalMeanReturn, {}, 'mean', {})

    def test_variance_update_moving_window(self):
        """Test the variance forecaster with moving window."""
        self._base_test_moving_window_vector_update(
            HistoricalVariance, {'kelly': False}, 'var', {'ddof': 0})

    def test_stddev_update_moving_window(self):
        """Test the standard deviation forecaster with moving window."""
        self._base_test_moving_window_vector_update(
            HistoricalStandardDeviation, {'kelly': False}, 'std', {'ddof': 0})

    def test_meanerror_update_moving_window(self):
        """Test standard deviation on the mean with moving window."""

        def _me(past_returns_noncash):
            return past_returns_noncash.std(ddof=0)\
                / np.sqrt(past_returns_noncash.count())

        self._base_test_moving_window_vector_update(
            HistoricalMeanError, {}, None, None, df_callable=_me)

    def test_cov_update_moving_window(self):
        """Test the covariance forecaster  with moving window."""

        self._base_test_moving_window_vector_update(
            HistoricalCovariance, {'kelly': True}, None, None,
            df_callable=self._kelly_covariance)

    def test_cov_update_moving_window_nokelly(self):
        """Test the covariance forecaster with moving window without Kelly."""

        self._base_test_moving_window_vector_update(
            HistoricalCovariance, {'kelly': False}, 'cov', {'ddof': 0},
            with_nans=False)

    def _base_test_exponential_moving_window_vector_update(
            self, forecaster, fc_kwargs, df_callable=None, with_nans=True):
        """Base test for vector quantities using an exponential moving window,
        and an exponential moving window in combination with a moving window."""
        half_life = pd.Timedelta('20d')
        inst_forecaster = forecaster(**fc_kwargs, half_life=half_life)
        returns = pd.DataFrame(self.returns, copy=True)
        if with_nans:
            returns.iloc[:40, 3:10] = np.nan

        for tidx in [50, 51, 52, 55, 56, 57, 58, 59, 60]:
            t = returns.index[tidx]
            past_returns = returns.loc[returns.index < t]

            cvx_val = inst_forecaster.values_in_time_recursive(
                t=t, past_returns=past_returns)

            self.assertFalse(np.any(np.isnan(cvx_val)))

            # print(cvx_val - df_callable(past_returns.iloc[:,:-1], half_life))
            self.assertTrue(np.allclose(
                cvx_val, df_callable(past_returns.iloc[:, :-1], half_life)))

        # Test with both MA and EMA

        half_life = pd.Timedelta('10d')
        window = pd.Timedelta('20d')
        inst_forecaster = forecaster(
            **fc_kwargs, rolling = window, half_life=half_life)
        returns = pd.DataFrame(self.returns, copy=True)
        if with_nans:
            returns.iloc[:40, 3:10] = np.nan

        for tidx in [50, 51, 52, 55, 56, 57, 58, 59, 60]:
            t = returns.index[tidx]
            past_returns = returns.loc[returns.index < t]
            past_returns_window = past_returns.loc[
                past_returns.index >= t - window]
            cvx_val = inst_forecaster.values_in_time_recursive(
                t=t, past_returns=past_returns)

            self.assertFalse(np.any(np.isnan(cvx_val)))
            self.assertTrue(np.allclose(
                cvx_val, df_callable(
                    past_returns_window.iloc[:, :-1], half_life)))

    @staticmethod
    def _mean_emw(past_returns_noncash, half_life):
        return past_returns_noncash.ewm(
            halflife=half_life, times=past_returns_noncash.index
            ).mean().iloc[-1]

    def test_mean_update_exponential_moving_window(self):
        """Test the mean forecaster with exponential moving window."""

        self._base_test_exponential_moving_window_vector_update(
            HistoricalMeanReturn, {}, self._mean_emw)

    @staticmethod
    def _var_emw(past_returns_noncash, half_life):
        """We need to do this b/c pandas.DataFrame.emw.var doesn't support
        the ddof=0 option."""
        return (past_returns_noncash**2).ewm(
                halflife=half_life, times=past_returns_noncash.index
                ).mean().iloc[-1] - TestForecast._mean_emw(
                    past_returns_noncash, half_life)**2

    def test_variance_update_exponential_moving_window(self):
        """Test the var forecaster with exponential moving window."""

        self._base_test_exponential_moving_window_vector_update(
            HistoricalVariance, {'kelly': False}, self._var_emw)

    def test_stddev_update_exponential_moving_window(self):
        """Test the std forecaster with exponential moving window."""
        def _std_emw(*args):
            return np.sqrt(self._var_emw(*args))
        self._base_test_exponential_moving_window_vector_update(
            HistoricalStandardDeviation, {'kelly': False}, _std_emw)

    def test_cov_update_exponential_moving_window(self):
        """Test the covariance forecaster with exponential moving window."""

        self._base_test_exponential_moving_window_vector_update(
            HistoricalCovariance, {'kelly': True},
            df_callable=self._kelly_covariance)

    def test_cov_update_exponential_moving_window_nokelly(self):
        """Test the covariance forecaster with exponential moving window."""

        self._base_test_exponential_moving_window_vector_update(
            HistoricalCovariance, {'kelly': False},
            df_callable=self._nokelly_covariance_emw_nonans, with_nans=False)

    def test_counts_matrix(self):
        """Test internal method(s) of HistoricalFactorizedCovariance."""
        forecaster = HistoricalFactorizedCovariance()  # kelly=True)
        returns = pd.DataFrame(self.returns, copy=True)
        returns.iloc[:20, 3:10] = np.nan
        returns.iloc[10:15, 10:20] = np.nan

        # pylint: disable=protected-access
        count_matrix = forecaster._compute_denominator(
            returns.iloc[:, :-1], None)

        for indexes in [(1, 2), (4, 5), (1, 5), (7, 18),
                (7, 24), (1, 15), (13, 22)]:
            print(count_matrix.iloc[indexes[0], indexes[1]])
            print(len((returns.iloc[:, indexes[0]] *
                  returns.iloc[:, indexes[1]]).dropna()))
            self.assertTrue(
                np.isclose(count_matrix.iloc[indexes[0], indexes[1]],
                    len((returns.iloc[:, indexes[0]]
                         * returns.iloc[:, indexes[1]]).dropna())))

    def test_sum_matrix(self):
        """Test internal method(s) of HistoricalFactorizedCovariance."""
        forecaster = HistoricalFactorizedCovariance()  # kelly=True)
        returns = pd.DataFrame(self.returns, copy=True)
        returns.iloc[:20, 3:10] = np.nan
        returns.iloc[10:15, 10:20] = np.nan

        forecaster.values_in_time_recursive(
            t=pd.Timestamp('2022-01-01'), past_returns=returns)

        # pylint: disable=protected-access
        sum_matrix = forecaster._numerator

        for indexes in [(1, 2), (4, 5), (1, 5), (7, 18),
                (7, 24), (1, 15), (13, 22)]:
            print()
            print(sum_matrix.iloc[indexes[0], indexes[1]])
            print((returns.iloc[:, indexes[0]] *
                  returns.iloc[:, indexes[1]]).sum())
            self.assertTrue(np.isclose(
                sum_matrix.iloc[indexes[0], indexes[1]],
                (returns.iloc[:, indexes[0]] *
                 returns.iloc[:, indexes[1]]).sum()
            ))

    def test_covariance_update_withnans(self):
        """Test covariance forecast estimator."""

        forecaster = HistoricalFactorizedCovariance()

        returns = pd.DataFrame(self.returns.iloc[:, :4], copy=True)
        returns.iloc[:20, 1] = np.nan
        returns.iloc[10:30, 0] = np.nan
        returns.iloc[25:40, 2] = np.nan

        def _compute_covariance(rets):
            res = np.zeros((3, 3))
            res[0, 0] = np.nanmean(rets.iloc[:, 0] * rets.iloc[:, 0])
            res[1, 1] = np.nanmean(rets.iloc[:, 1] * rets.iloc[:, 1])
            res[2, 2] = np.nanmean(rets.iloc[:, 2] * rets.iloc[:, 2])
            res[0, 1] = np.nanmean(rets.iloc[:, 0] * rets.iloc[:, 1])
            res[0, 2] = np.nanmean(rets.iloc[:, 0] * rets.iloc[:, 2])
            res[1, 2] = np.nanmean(rets.iloc[:, 1] * rets.iloc[:, 2])
            res[1, 0] = res[0, 1]
            res[2, 0] = res[0, 2]
            res[2, 1] = res[1, 2]
            return res

        for tidx in [50, 51, 52, 55, 56, 57]:
            t = returns.index[tidx]
            past_returns = returns.loc[returns.index < t]
            val = forecaster.values_in_time_recursive(
                t=t, past_returns=past_returns)
            Sigma = val @ val.T
            self.assertTrue(
                np.allclose(Sigma, _compute_covariance(past_returns)))

    def test_covariance_update_nokelly_withnans(self):
        """Test covariance forecast estimator.

        NOTE: due to a bug in pandas we can't test against
        pandas.DataFrame.cov,
        see https://github.com/pandas-dev/pandas/issues/45814 .
        In fact with the current bug in pandas
        ``past_returns.iloc[:,:-1].cov(ddof=0)`` returns
        ``past_returns.iloc[:,:-1].cov(ddof=1)``
        whenever there are missing values.
        """

        forecaster = HistoricalFactorizedCovariance(kelly=False)

        returns = pd.DataFrame(self.returns.iloc[:, :4], copy=True)
        returns.iloc[:20, 1] = np.nan
        returns.iloc[10:30, 0] = np.nan
        returns.iloc[25:40, 2] = np.nan

        def _cov_ij(i, j, rets):
            i_nanmasker = np.zeros(len(rets))
            i_nanmasker[rets.iloc[:, i].isnull()] = np.nan
            i_nanmasker[~(rets.iloc[:, i].isnull())] = 1.
            j_nanmasker = np.zeros(len(rets))
            j_nanmasker[rets.iloc[:, j].isnull()] = np.nan
            j_nanmasker[~(rets.iloc[:, j].isnull())] = 1.
            print(i_nanmasker, j_nanmasker)
            return np.nanmean(rets.iloc[:, i] * rets.iloc[:, j]
                ) - np.nanmean(rets.iloc[:, i] * j_nanmasker
                    ) * np.nanmean(rets.iloc[:, j] * i_nanmasker)

        def _compute_covariance(rets):
            res = np.zeros((3, 3))
            res[0, 0] = _cov_ij(0, 0, rets)
            res[1, 1] = _cov_ij(1, 1, rets)
            res[2, 2] = _cov_ij(2, 2, rets)
            res[0, 1] = _cov_ij(0, 1, rets)
            res[0, 2] = _cov_ij(0, 2, rets)
            res[1, 2] = _cov_ij(1, 2, rets)
            res[1, 0] = res[0, 1]
            res[2, 0] = res[0, 2]
            res[2, 1] = res[1, 2]
            return res

        for tidx in [50, 51, 52, 55, 56, 57]:
            t = returns.index[tidx]
            past_returns = returns.loc[returns.index < t]
            val = forecaster.values_in_time_recursive(
                t=t, past_returns=past_returns)
            Sigma = val @ val.T

            self.assertTrue(
                np.allclose(Sigma, _compute_covariance(past_returns)))
            # pandasSigma = past_returns.iloc[:,:-1].cov(ddof=0)
            # self.assertTrue(np.allclose(Sigma, pandasSigma))
            self.assertTrue(np.allclose(
                np.diag(Sigma), past_returns.iloc[:, :-1].var(ddof=0)))

    def test_svd_forecaster(self):
        """Test the SVD forecaster.

        In particular, we compare it with the other covariance
        forecaster.

        We check that their forecasts are the same if there are no
        missing values in the past returns, and they diverge more as
        more missing values are introduced. Additionally, we check that
        it breaks when there are too many (more than 90%!) missing
        values.
        """

        returns = pd.DataFrame(self.returns.iloc[:, :4], copy=True)
        t = returns.index[10]

        num_factors = 1

        forecaster = HistoricalLowRankCovarianceSVD(
            num_factors=num_factors)
        forecaster2 = HistoricalFactorizedCovariance(kelly=True)

        def _compare_with_eigh(returns):

            F, d = forecaster.values_in_time_recursive(
                t=t, past_returns=returns)

            sigma_fact = forecaster2.values_in_time_recursive(
                t=t, past_returns=returns)

            sigma_svd = F.T @ F + np.diag(d)
            sigma_eigh = (
                sigma_fact[:, -num_factors:] @ sigma_fact[:, -num_factors:].T
                + np.diag((sigma_fact[:, :-num_factors]**2).sum(1)))
            print(sigma_svd)
            print(sigma_eigh)
            print(np.linalg.norm(sigma_svd),
                  np.linalg.norm(sigma_eigh),
                  np.linalg.norm(sigma_eigh-sigma_svd)
                  )

            forecaster3 = HistoricalLowRankCovarianceSVD(num_factors=1,
                svd='scipy')

            with self.assertRaises(SyntaxError):
                forecaster3.values_in_time_recursive(t=t, past_returns=returns)

            return np.linalg.norm(sigma_eigh-sigma_svd)

        self.assertTrue(np.isclose(_compare_with_eigh(returns), 0.))

        returns.iloc[:20, 1] = np.nan

        diff1 = _compare_with_eigh(returns)
        self.assertTrue(0 < diff1)

        returns.iloc[10:30, 0] = np.nan

        diff2 = _compare_with_eigh(returns)
        self.assertTrue(diff1 < diff2)

        returns.iloc[25:40, 2] = np.nan
        diff3 = _compare_with_eigh(returns)
        self.assertTrue(diff2 < diff3)

        print(returns.isnull().mean())
        returns.iloc[4:-3, -2] = np.nan
        print(returns.isnull().mean())
        diff4 = _compare_with_eigh(returns)
        self.assertTrue(diff3 < diff4)

        returns.iloc[:50, 0] = np.nan
        returns.iloc[50:, 1] = np.nan
        print(returns.isnull().mean())

        with self.assertRaises(ForecastError):
            forecaster2.values_in_time_recursive(t=t, past_returns=returns)

        # this one is even more robust!
        forecaster.values_in_time_recursive(
            t=t, past_returns=returns)

        returns.iloc[56:, 0] = np.nan
        print(returns.isnull().mean())

        forecaster.values_in_time_recursive(
            t=t, past_returns=returns)

        returns.iloc[:70, 1] = np.nan
        print(returns.isnull().mean())

        with self.assertRaises(ForecastError):
            forecaster.values_in_time_recursive(
                t=t, past_returns=returns)


if __name__ == '__main__':

    unittest.main() # pragma: no cover



# Copyright 2016 Enzo Busseti, Stephen Boyd, Steven Diamond, BlackRock Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Unit tests for the risk objects."""

import unittest

import numpy as np
import pandas as pd

import cvxportfolio as cvx
from cvxportfolio.tests import CvxportfolioTest

USED_RETURNS = 10


class TestRisks(CvxportfolioTest):
    """Test risk objects."""

    def test_full_sigma(self):
        """Test full covariance model."""
        historical_covariances = self.returns.iloc[
            :, :-1].rolling(50).cov(ddof=0).dropna()
        risk_model = cvx.FullCovariance(historical_covariances)

        cvxpy_expression = self.boilerplate(risk_model)

        self.assertTrue(cvxpy_expression.is_convex())

        t = historical_covariances.index[123][0]
        self.w_plus_minus_w_bm.value = np.random.randn(self.N)

        risk_model.values_in_time_recursive(t=t, past_returns='Hello!')

        self.assertTrue(np.isclose(
            cvxpy_expression.value,
            self.w_plus_minus_w_bm.value[:-1] @ historical_covariances.loc[t]
                @ self.w_plus_minus_w_bm.value[:-1]))

    def test_full_estimated_sigma(self):
        """Test full covariance model estimated internally."""
        risk_model = cvx.FullCovariance()

        cvxpy_expression = self.boilerplate(risk_model)
        self.assertTrue(cvxpy_expression.is_convex())

        t = pd.Timestamp('2014-06-02')
        past = self.returns.loc[self.returns.index < t]
        should_be = past.iloc[:, :-1].T @ past.iloc[:, :-1] / len(past)

        self.w_plus_minus_w_bm.value = np.random.randn(self.N)

        risk_model.values_in_time_recursive(t=t, past_returns=past)
        print(cvxpy_expression.value)

        print(self.w_plus_minus_w_bm.value[:-1] @
              should_be @ self.w_plus_minus_w_bm.value[:-1])

        self.assertTrue(
            np.isclose(cvxpy_expression.value,
            self.w_plus_minus_w_bm.value[:-1] @ should_be
                @ self.w_plus_minus_w_bm.value[:-1]))

    def test_diagonal_covariance(self):
        """Test diagonal covariance model."""
        historical_variances = self.returns.iloc[
            :, :-1].rolling(50).var().shift(1).dropna()
        risk_model = cvx.DiagonalCovariance(historical_variances)
        cvxpy_expression = self.boilerplate(risk_model)
        self.assertTrue(cvxpy_expression.is_convex())

        t = historical_variances.index[123]
        self.w_plus_minus_w_bm.value = np.random.randn(self.N)

        risk_model.values_in_time_recursive(t=t, past_returns='hello')

        self.assertTrue(
            np.isclose(cvxpy_expression.value,
                self.w_plus_minus_w_bm.value[:-1]
                @ np.diag(historical_variances.loc[t])
                @ self.w_plus_minus_w_bm.value[:-1]))

    def test_full_diagonal_covariance(self):
        """Test diagonal covariance estimated internally."""

        risk_model = cvx.DiagonalCovariance()
        cvxpy_expression = self.boilerplate(risk_model)
        self.assertTrue(cvxpy_expression.is_convex())

        t = pd.Timestamp('2014-06-02')
        past = self.returns.loc[self.returns.index < t]
        should_be = (past**2).mean()
        should_be.iloc[-1] = 0.

        risk_model.values_in_time_recursive(t=t, past_returns=past)
        self.w_plus_minus_w_bm.value = np.random.randn(self.N)

        self.assertTrue(
            np.isclose(cvxpy_expression.value, self.w_plus_minus_w_bm.value
                @ np.diag(should_be) @ self.w_plus_minus_w_bm.value))

    def test_forecast_error(self):
        """Test RiskForecastError term."""

        historical_variances = (
            self.returns.iloc[:, :-1]**2).rolling(50).mean().shift(1).dropna()

        risk_model = cvx.RiskForecastError(historical_variances)
        cvxpy_expression = self.boilerplate(risk_model)

        self.assertTrue(cvxpy_expression.is_convex())

        t = historical_variances.index[123]

        self.w_plus_minus_w_bm.value = np.random.randn(self.N)

        risk_model.values_in_time_recursive(t=t, past_returns='hello')

        print(cvxpy_expression.value)
        print((np.abs(
            self.w_plus_minus_w_bm.value[:-1])
                @ np.sqrt(historical_variances.loc[t]))**2)
        self.assertTrue(
            np.isclose(cvxpy_expression.value,
            (np.abs(self.w_plus_minus_w_bm.value[:-1])
                @ np.sqrt(historical_variances.loc[t]))**2))

    def test_full_forecast_error(self):
        """Test RiskForecastError term estimated internally."""

        risk_model = cvx.RiskForecastError()
        cvxpy_expression = self.boilerplate(risk_model)
        self.assertTrue(cvxpy_expression.is_convex())

        t = pd.Timestamp('2014-06-02')

        past = self.returns.loc[self.returns.index < t]
        should_be = (past**2).mean()
        should_be.iloc[-1] = 0.

        risk_model.values_in_time_recursive(t=t, past_returns=past)
        self.w_plus_minus_w_bm.value = np.random.randn(self.N)

        self.assertTrue(
            np.isclose(cvxpy_expression.value,
            (np.abs(self.w_plus_minus_w_bm.value) @ np.sqrt(should_be))**2 ))

    def test_low_rank_covariance(self):
        """Test FactorModelCovariance."""

        F = pd.DataFrame(np.random.randn(2, self.N-1),
                         columns=self.returns.columns[:-1])
        d = pd.Series(np.random.uniform(size=self.N-1),
                      self.returns.columns[:-1])
        risk_model = cvx.FactorModelCovariance(F=F, d=d)

        cvxpy_expression = self.boilerplate(risk_model)
        self.assertTrue(cvxpy_expression.is_convex())

        risk_model.values_in_time_recursive(
            t=self.returns.index[12], past_returns='hello')
        self.w_plus_minus_w_bm.value = np.random.randn(self.N)

        self.assertTrue(
            np.isclose(cvxpy_expression.value,
            self.w_plus_minus_w_bm.value[:-1] @ np.diag(d)
                @ self.w_plus_minus_w_bm.value[:-1]
                + ((F @ self.w_plus_minus_w_bm.value[:-1])**2).sum()))

    def test_low_rank_covariance_with_factors_covariance(self):
        """Test FactorModelCovariance with non-identity factors covariance."""

        F = pd.DataFrame(np.random.randn(2, self.N-1),
                         columns=self.returns.columns[:-1])
        d = pd.Series(np.random.uniform(size=self.N-1),
                      self.returns.columns[:-1])
        _ = np.random.randn(2, 2)
        factors_cov = _.T @ _
        risk_model = cvx.FactorModelCovariance(F=F, d=d, Sigma_F=factors_cov)

        cvxpy_expression = self.boilerplate(risk_model)
        self.assertTrue(cvxpy_expression.is_convex())

        risk_model.values_in_time_recursive(
            t=self.returns.index[12], past_returns='hello')
        self.w_plus_minus_w_bm.value = np.random.randn(self.N)

        self.assertTrue(
            np.isclose(cvxpy_expression.value,
                self.w_plus_minus_w_bm.value[:-1] @ np.diag(d)
                    @ self.w_plus_minus_w_bm.value[:-1]
                    + self.w_plus_minus_w_bm.value[:-1].T @ F.T
                    @ factors_cov @ F @ self.w_plus_minus_w_bm.value[:-1]))

    def test_estimated_low_rank_covariance(self):
        """Test FactorModelCovariance estimated internally."""

        risk_model = cvx.FactorModelCovariance()

        cvxpy_expression = self.boilerplate(risk_model)
        self.assertTrue(cvxpy_expression.is_convex())

        t = pd.Timestamp('2014-06-02')

        past = self.returns.loc[self.returns.index < t]

        fullsigma = past.T @ past / len(past)
        fullsigma.iloc[:, -1] = 0.
        fullsigma.iloc[-1, :] = 0.
        eigval, eigvec = np.linalg.eigh(fullsigma)
        F = np.sqrt(eigval[-1]) * eigvec[:, -1]
        d = np.diag(fullsigma) - F**2

        risk_model.values_in_time_recursive(t=t, past_returns=past)
        self.w_plus_minus_w_bm.value = np.random.randn(self.N)

        self.assertTrue(
            np.isclose(cvxpy_expression.value, self.w_plus_minus_w_bm.value
                @ np.diag(d) @ self.w_plus_minus_w_bm.value
                + ((F @ self.w_plus_minus_w_bm.value)**2).sum()))

    def test_worst_case_risk(self):
        """Test the worst-case risk object."""

        risk_model0 = cvx.FullCovariance()
        risk_model1 = cvx.DiagonalCovariance()
        worst_case = cvx.WorstCaseRisk([risk_model0, risk_model1])

        cvxpy_expression = self.boilerplate(worst_case)

        assert cvxpy_expression.is_convex()

        cvxpy_expression0 = risk_model0.compile_to_cvxpy(
            self.w_plus, self.z, self.w_plus_minus_w_bm)
        cvxpy_expression1 = risk_model1.compile_to_cvxpy(
            self.w_plus, self.z, self.w_plus_minus_w_bm)

        self.w_plus_minus_w_bm.value = np.ones(self.N)

        t = pd.Timestamp('2014-06-02')

        worst_case.values_in_time_recursive(
            t=t, past_returns=self.returns.loc[self.returns.index < t])

        print(cvxpy_expression.value)
        print(cvxpy_expression0.value)
        print(cvxpy_expression1.value)
        self.assertTrue(cvxpy_expression.value == cvxpy_expression0.value)
        self.assertTrue(cvxpy_expression.value > cvxpy_expression1.value)


if __name__ == '__main__':

    unittest.main() # pragma: no cover



# Copyright 2023 Enzo Busseti
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module include classes that download, store, and serve market data.

The two main abstractions are :class:`SymbolData` and :class:`MarketData`.
Neither are exposed outside this module. Their derived classes instead are.
If you want to interface cvxportfolio with financial data source other
than the ones we provide, you should derive from either of those two classes.
"""

from .market_data import *
from .symbol_data import *

__all__ = [
    "YahooFinance", "Fred", "UserProvidedMarketData", "DownloadedMarketData"]



# Copyright 2016 Enzo Busseti, Stephen Boyd, Steven Diamond, BlackRock Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Unit tests for some utils.py functions."""

import unittest

import numpy as np
import pandas as pd

from cvxportfolio.utils import make_numeric, repr_numpy_pandas


class TestUtils(unittest.TestCase):
    """Test functions in the cvxportfolio.utils module."""

    def test_hasher(self):
        """Test hashing function."""
        a = np.array([1, 2, 3])
        re = repr_numpy_pandas(a)
        print(re)
        b = np.array([[1, 2, 3]])
        re1 = repr_numpy_pandas(b)
        print(re1)
        self.assertTrue(re == re1)

        with self.assertRaises(NotImplementedError):
            repr_numpy_pandas('Hello!')

        a = pd.Series([1, 2, 3], ['2020-01-01', '2021-01-01', '2022-01-01'])
        re = repr_numpy_pandas(a)
        print(re)
        b = pd.Series(a.values, a.index)
        re1 = repr_numpy_pandas(b)
        print(re1)
        self.assertTrue(re == re1)

        # multiindexed df
        timeindex = pd.date_range("2022-01-01", "2022-01-30")
        second_level = ["hello", "ciao", "hola"]
        index = pd.MultiIndex.from_product([timeindex, second_level])
        data = pd.DataFrame(np.random.randn(len(index), 10), index=index)
        data.columns = ["one", "two", "tre", "quattro",
                        "cinque", "sei", "sette", "otto", "nove", "dieci"]

        re = repr_numpy_pandas(data)
        print(re)

        data1 = pd.DataFrame(data.values,
                             index=data.index, columns=data.columns)

        re1 = repr_numpy_pandas(data1)
        print(re1)
        self.assertTrue(re == re1)

    def test_make_numeric(self):
        """Test make_numeric function for user-provided data."""

        for data in [
                2., np.array([1, 2]), pd.Series([1, 2, 3]),
                pd.DataFrame([[1, 2., 3], [4, 5., 6]])]:
            self.assertTrue(make_numeric(data) is data)

        for data in [
                np.array([1, 2], dtype=object), pd.Series([1, 2, 3], dtype=object),
                pd.DataFrame([[1, 2., 3], [4, 5., 6]], dtype=object)]:
            self.assertTrue(np.all(make_numeric(data) == data))

        for data in [
                np.array(['1', 2], dtype=object),
                pd.Series([1, '2', 3], dtype=object),
                pd.DataFrame([[1, '2.', 3], [4, '5.', 6]], dtype=object)]:
            self.assertTrue(np.all(data.astype(float) == make_numeric(data)))

        for data in [
                np.array(['1a', 2], dtype=object),
                pd.Series([1, '2a', 3], dtype=object),
                pd.DataFrame([[1, '2a.', 3], [4, '5a.', 6]], dtype=object)]:
            with self.assertRaises(ValueError):
                make_numeric(data)


if __name__ == '__main__':

    unittest.main() # pragma: no cover
